# Protean
from protean.core.aggregate import BaseAggregate
from protean.core.entity import BaseEntity
from protean.core.field.basic import String
from protean.utils import fully_qualified_name


class TestEntityRegistration:
    def test_manual_registration_of_entity(self, test_domain):
        class Post(BaseAggregate):
            name = String(max_length=50)

        class Comment(BaseEntity):
            content = String(max_length=500)

            class Meta:
                aggregate_cls = Post

        test_domain.register(Post)
        test_domain.register(Comment)

        assert fully_qualified_name(Comment) in test_domain.registry.entities
        assert Comment.meta_.aggregate_cls == Post

    def test_setting_provider_in_decorator_based_registration(self, test_domain):
        @test_domain.aggregate
        class Post:
            name = String(max_length=50)

        @test_domain.entity
        class Comment(BaseEntity):
            content = String(max_length=500)

            class Meta:
                aggregate_cls = Post

        assert Comment.meta_.aggregate_cls == Post

    def test_setting_provider_in_decorator_based_registration_with_parameters(
        self, test_domain
    ):
        @test_domain.aggregate
        class Post:
            name = String(max_length=50)

        @test_domain.entity(aggregate_cls=Post)
        class Comment(BaseEntity):
            content = String(max_length=500)

        assert Comment.meta_.aggregate_cls == Post

    def test_register_entity_against_a_dummy_aggregate(self, test_domain):
        # Though the registration succeeds, this will eventually fail
        #   when the domain tries to resolve the aggregate.
        # Protean
        from protean.core.field.basic import String

        @test_domain.entity(aggregate_cls="foo")
        class FooBar:
            foo = String(max_length=50)

        assert FooBar.meta_.aggregate_cls == "foo"



""" Test cases for all available field type implementations"""

# Standard Library Imports
import enum

from datetime import datetime

# Protean
import pytest

from protean.core.aggregate import BaseAggregate
from protean.core.exceptions import ValidationError
from protean.core.field.basic import (
    Auto,
    Boolean,
    Date,
    DateTime,
    Dict,
    Float,
    Integer,
    List,
    String,
    Text,
)


class TestStringField:
    """ Test the String Field Implementation"""

    def test_init(self):
        """Test successful String Field initialization"""

        name = String(max_length=10)
        assert name is not None

    def test_type_validation(self):
        """ Test type checking validation for the Field"""
        name = String(max_length=10)
        assert name._load(1) == "1"

    def test_min_length(self):
        """ Test minimum length validation for the string field"""

        with pytest.raises(ValidationError):
            name = String(min_length=5, max_length=10)
            name._load("Dum")

    def test_max_length(self):
        """ Test maximum length validation for the string field"""

        with pytest.raises(ValidationError):
            name = String(max_length=5)
            name._load("Dummy Dummy")

    def test_choice(self):
        """ Test choices validations for the string field """

        class StatusChoices(enum.Enum):
            """ Set of choices for the status"""

            PENDING = "Pending"
            SUCCESS = "Success"
            ERROR = "Error"

        status = String(max_length=10, choices=StatusChoices)
        assert status is not None

        # Test loading of values to the status field
        assert status._load("Pending") == "Pending"
        with pytest.raises(ValidationError) as e_info:
            status._load("Failure")
        assert e_info.value.messages == {
            "unlinked": [
                "Value `'Failure'` is not a valid choice. "
                "Must be one of ['Pending', 'Success', 'Error']"
            ]
        }


class TestIntegerField:
    """ Test the Integer Field Implementation"""

    def test_init(self):
        """Test successful Integer Field initialization"""

        age = Integer()
        assert age is not None

    def test_various_input_values(self):
        age = Integer()
        assert age._load(12) == 12
        assert age._load("12") == 12
        assert age._load(None) is None
        assert age._load("") is None

    def test_type_validation(self):
        """ Test type checking validation for the Field"""
        with pytest.raises(ValidationError):
            age = Integer()
            age._load("x")

    def test_min_value(self):
        """ Test minimum value validation for the integer field"""

        with pytest.raises(ValidationError):
            age = Integer(min_value=5)
            age._load(3)

    def test_max_value(self):
        """ Test maximum value validation for the integer field"""

        with pytest.raises(ValidationError):
            age = Integer(max_value=5)
            age._load(6)

    def test_choice(self):
        """ Test choices validations for the Integer field """

        class StatusChoices(enum.Enum):
            """ Set of choices for the status"""

            PENDING = (0, "Pending")
            SUCCESS = (1, "Success")
            ERROR = (2, "Error")

        status = Integer(choices=StatusChoices)
        assert status is not None

        # Test loading of values to the status field
        assert status._load(0) == 0
        with pytest.raises(ValidationError) as e_info:
            status._load(4)
        assert e_info.value.messages == {
            "unlinked": ["Value `4` is not a valid choice. " "Must be one of [0, 1, 2]"]
        }


class TestFloatField:
    """ Test the Float Field Implementation"""

    def test_init(self):
        """Test successful Float Field initialization"""

        score = Float()
        assert score is not None

    def test_type_validation(self):
        """ Test type checking validation for the Field"""
        with pytest.raises(ValidationError):
            score = Float()
            score._load("x")

    def test_min_value(self):
        """ Test minimum value validation for the float field"""

        with pytest.raises(ValidationError):
            score = Float(min_value=5.4)
            score._load(5.3)

    def test_max_value(self):
        """ Test maximum value validation for the float field"""

        with pytest.raises(ValidationError):
            score = Float(max_value=5.5)
            score._load(5.6)

    @pytest.mark.xfail
    def test_none_value(self):
        """ Test None value treatment for the float field"""

        score = Float(max_value=5.5)
        score._load(None)

        assert score.value == 0.0


class TestBooleanField:
    """ Test the Boolean Field Implementation"""

    def test_init(self):
        """Test successful Boolean Field initialization"""

        married = Boolean()
        assert married is not None
        assert married._load(True) is True

    def test_type_validation(self):
        """ Test type checking validation for the Field"""
        with pytest.raises(ValidationError):
            married = Boolean()
            married._load("x")

    def test_default_value(self):
        """Test that Boolean fields accept default values properly"""

        class Youth(BaseAggregate):
            name = String(max_length=50)
            married = Boolean(default=False)

        class Adult(BaseAggregate):
            name = String(max_length=50)
            married = Boolean(default=True)

        youth = Youth(name="Baby Doe")
        adult = Adult(name="John Doe")

        assert youth.married is False
        assert adult.married is True


class TestListField:
    """ Test the List Field Implementation"""

    def test_init(self):
        """Test successful List Field initialization"""

        tags = List()
        assert tags is not None

        assert tags._load(["x", "y", "z"]) == ["x", "y", "z"]

    def test_type_validation(self):
        """ Test type checking validation for the Field"""
        numbers = List(content_type=Integer)
        with pytest.raises(ValidationError):
            numbers._load("x")

    def test_choice(self):
        """ Test choices validations for the list field """

        class StatusChoices(enum.Enum):
            """ Set of choices for the status"""

            PENDING = "Pending"
            SUCCESS = "Success"
            ERROR = "Error"

        status = List(choices=StatusChoices)
        assert status is not None

        # Test loading of values to the status field
        assert status._load(["Pending"]) == ["Pending"]
        with pytest.raises(ValidationError) as e_info:
            status._load(["Pending", "Failure"])
        assert e_info.value.messages == {
            "unlinked": [
                "Value `'Failure'` is not a valid choice. "
                "Must be one of ['Pending', 'Success', 'Error']"
            ]
        }


class TestDictField:
    """ Test the Dict Field Implementation"""

    def test_init(self):
        """Test successful Dict Field initialization"""

        add_info = Dict()
        assert add_info is not None

        value = add_info._load({"available": "weekdays"})
        assert value == {"available": "weekdays"}

    def test_type_validation(self):
        """ Test type checking validation for the Field"""
        with pytest.raises(ValidationError):
            add_info = Dict()
            add_info._load("x")


class TestAutoField:
    """ Test the Auto Field Implementation"""

    def test_init(self):
        """Test successful Dict Field initialization"""

        add_info = Auto()
        assert add_info is not None

        value = add_info._load(1)
        assert value == 1

    def test_validation(self):
        """ Test validation for the Auto Field"""
        add_info = Auto(required=True)
        add_info._load(None)


class TestDateField:
    """ Test the Date Field Implementation"""

    def test_init(self):
        """Test successful Date Field initialization"""

        age = Date()
        assert age is not None

        value = age._load(datetime.now().date())
        assert value == datetime.now().date()

    def test_type_casting(self):
        """ Test type casting and validation for the Field"""

        age = Date()

        # Test datetime being passed as value
        assert age._load(datetime.now()) == datetime.now().date()

        # Test string dates being passed as value
        expected = datetime(2018, 3, 16).date()
        assert age._load("2018-03-16") == expected
        assert age._load("2018-03-16 10:23:32") == expected
        assert age._load("16th March 2018") == expected

        # Test for invalid date
        with pytest.raises(ValidationError):
            assert age._load("15 Marchs")


class TestDateTimeField:
    """ Test the DateTime Field Implementation"""

    def test_init(self):
        """Test successful DateTime Field initialization"""

        created_at = DateTime()
        assert created_at is not None

        value = datetime.now()
        assert value == created_at._load(value)

    def test_type_casting(self):
        """ Test type casting and validation for the Field"""

        created_at = DateTime()
        today = datetime.now()
        # Test date being passed as value
        assert created_at._load(today.date()) == datetime(
            today.year, today.month, today.day
        )

        # Test string dates being passed as value
        assert created_at._load("2018-03-16") == datetime(2018, 3, 16)
        assert created_at._load("2018-03-16 10:23:32") == datetime(
            2018, 3, 16, 10, 23, 32
        )

        # Test for invalid datetime
        with pytest.raises(ValidationError):
            assert created_at._load("2018-03-16 10 23 32")


class TestTextField:
    """ Test the Text Field Implementation"""

    def test_init(self):
        """Test successful Text Field initialization"""

        address = Text()
        assert address is not None

    def test_type_validation(self):
        """ Test type checking validation for the Field"""
        address = Text()
        value = address._load("My home address")
        assert value == "My home address"



"""Module with repository implementation for SQLAlchemy"""
# Standard Library Imports
import logging
import uuid

from abc import ABCMeta
from typing import Any

# Protean
from protean.core.exceptions import ConfigurationError, ObjectNotFoundError
from protean.core.field.association import Reference
from protean.core.field.basic import (
    Auto,
    Boolean,
    Date,
    DateTime,
    Dict,
    Float,
    Identifier,
    Integer,
    List,
    String,
    Text,
)
from protean.core.model import BaseModel
from protean.globals import current_domain, current_uow
from protean.port.dao import BaseDAO, BaseLookup, ResultSet
from protean.port.provider import BaseProvider
from protean.utils import Database, IdentityType
from protean.utils.query import Q
from sqlalchemy import Column, MetaData, and_, create_engine, or_, orm
from sqlalchemy import types as sa_types
from sqlalchemy.dialects.postgresql import ARRAY, JSON, UUID
from sqlalchemy.engine.url import make_url
from sqlalchemy.exc import DatabaseError
from sqlalchemy.ext import declarative as sa_dec
from sqlalchemy.ext.declarative import as_declarative, declared_attr
from sqlalchemy.types import CHAR, TypeDecorator

logging.getLogger("sqlalchemy").setLevel(logging.ERROR)
logger = logging.getLogger("protean.repository")


class GUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    """

    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value


def _get_identity_type():
    """Retrieve the configured data type for AutoGenerated Identifiers

    If `current_domain` is not yet available, it simply means that Protean is still being loaded.
    Default to `Identity.STRING`
    """
    try:
        if current_domain.config["IDENTITY_TYPE"] == IdentityType.INTEGER:
            return sa_types.Integer
        elif current_domain.config["IDENTITY_TYPE"] == IdentityType.STRING:
            return sa_types.String
        elif current_domain.config["IDENTITY_TYPE"] == IdentityType.UUID:
            return GUID
        else:
            raise ConfigurationError(
                f'Unknown Identity Type {current_domain.config["IDENTITY_TYPE"]}'
            )
    except RuntimeError as exc:
        logger.error(f"RuntimeError while identifying data type for identities: {exc}")
        return sa_types.String


class DeclarativeMeta(sa_dec.DeclarativeMeta, ABCMeta):
    """ Metaclass for the Sqlalchemy declarative schema """

    def __init__(cls, classname, bases, dict_):
        # Update the class attrs with the entity attributes

        field_mapping = {
            Auto: _get_identity_type(),
            Boolean: sa_types.Boolean,
            Date: sa_types.Date,
            DateTime: sa_types.DateTime,
            Dict: sa_types.PickleType,
            Float: sa_types.Float,
            Identifier: _get_identity_type(),
            Integer: sa_types.Integer,
            List: sa_types.PickleType,
            String: sa_types.String,
            Text: sa_types.Text,
        }

        if "meta_" in dict_:
            entity_cls = dict_["meta_"].entity_cls
            for _, field_obj in entity_cls.meta_.attributes.items():
                attribute_name = field_obj.attribute_name

                # Map the field if not in attributes
                if attribute_name not in cls.__dict__:
                    field_cls = type(field_obj)
                    type_args = []
                    type_kwargs = {}

                    # Get the SA type
                    sa_type_cls = field_mapping.get(field_cls)

                    # Upgrade to Postgresql specific Data Types
                    if cls.metadata.bind.dialect.name == "postgresql":

                        if field_cls == Dict and not field_obj.pickled:
                            sa_type_cls = JSON

                        if field_cls == List and not field_obj.pickled:
                            sa_type_cls = ARRAY

                            # Associate Content Type
                            if field_obj.content_type:
                                type_args.append(
                                    field_mapping.get(field_obj.content_type)
                                )
                            else:
                                type_args.append(sa_types.Text)

                    # Default to the text type if no mapping is found
                    if not sa_type_cls:
                        sa_type_cls = sa_types.String

                    # Build the column arguments
                    col_args = {
                        "primary_key": field_obj.identifier,
                        "nullable": not field_obj.required,
                        "unique": field_obj.unique,
                    }

                    # Update the arguments based on the field type
                    if issubclass(field_cls, String):
                        type_kwargs["length"] = field_obj.max_length

                    # Update the attributes of the class
                    setattr(
                        cls,
                        attribute_name,
                        Column(sa_type_cls(*type_args, **type_kwargs), **col_args),
                    )
        super().__init__(classname, bases, dict_)


def derive_schema_name(model_cls):
    if hasattr(model_cls.meta_, "schema_name"):
        return model_cls.meta_.schema_name
    else:
        return model_cls.meta_.entity_cls.meta_.schema_name


@as_declarative(metaclass=DeclarativeMeta)
class SqlalchemyModel(BaseModel):
    """Model representation for the Sqlalchemy Database """

    @declared_attr
    def __tablename__(cls):
        return derive_schema_name(cls)

    @classmethod
    def from_entity(cls, entity):
        """ Convert the entity to a model object """
        item_dict = {}
        for attribute_obj in cls.meta_.entity_cls.meta_.attributes.values():
            if isinstance(attribute_obj, Reference):
                item_dict[
                    attribute_obj.relation.attribute_name
                ] = attribute_obj.relation.value
            else:
                item_dict[attribute_obj.attribute_name] = getattr(
                    entity, attribute_obj.attribute_name
                )
        return cls(**item_dict)

    @classmethod
    def to_entity(cls, model_obj: "SqlalchemyModel"):
        """ Convert the model object to an entity """
        item_dict = {}
        for field_name in cls.meta_.entity_cls.meta_.attributes:
            item_dict[field_name] = getattr(model_obj, field_name, None)
        return cls.meta_.entity_cls(item_dict, raise_errors=False)


class SADAO(BaseDAO):
    """DAO implementation for Databases compliant with SQLAlchemy"""

    def __repr__(self) -> str:
        return f"SQLAlchemyDAO <{self.entity_cls.__name__}>"

    def _get_session(self):
        """Returns an active connection to the persistence store.

        - If there is an active transaction, the connection associated with the transaction (in the UoW) is returned
        - If the DAO has been explicitly instructed to work outside a UoW (with the help of `_outside_uow`), or if
            there are no active transactions, a new connection is retrieved from the provider and returned.

        Overridden here instead of using the version in `BaseDAO` because the connection needs to be started
            with a call to `begin()` if it is not yet active (checked with `is_active`)
        """
        if current_uow and not self._outside_uow:
            return current_uow.get_session(self.provider.name)
        else:
            new_connection = self.provider.get_connection()
            if not new_connection.is_active:
                new_connection.begin()
            return new_connection

    def _build_filters(self, criteria: Q):
        """ Recursively Build the filters from the criteria object"""
        # Decide the function based on the connector type
        func = and_ if criteria.connector == criteria.AND else or_
        params = []
        for child in criteria.children:
            if isinstance(child, Q):
                # Call the function again with the child
                params.append(self._build_filters(child))
            else:
                # Find the lookup class and the key
                stripped_key, lookup_class = self.provider._extract_lookup(child[0])

                # Instantiate the lookup class and get the expression
                lookup = lookup_class(stripped_key, child[1], self.model_cls)
                if criteria.negated:
                    params.append(~lookup.as_expression())
                else:
                    params.append(lookup.as_expression())

        return func(*params)

    def _filter(
        self, criteria: Q, offset: int = 0, limit: int = 10, order_by: list = ()
    ) -> ResultSet:
        """ Filter objects from the sqlalchemy database """
        conn = self._get_session()
        qs = conn.query(self.model_cls)

        # Build the filters from the criteria
        if criteria.children:
            qs = qs.filter(self._build_filters(criteria))

        # Apply the order by clause if present
        order_cols = []
        for order_col in order_by:
            col = getattr(self.model_cls, order_col.lstrip("-"))
            if order_col.startswith("-"):
                order_cols.append(col.desc())
            else:
                order_cols.append(col)
        qs = qs.order_by(*order_cols)
        qs_without_limit = qs
        qs = qs.limit(limit).offset(offset)

        # Return the results
        try:
            items = qs.all()
            result = ResultSet(
                offset=offset, limit=limit, total=qs_without_limit.count(), items=items
            )
        except DatabaseError as exc:
            logger.error(f"Error while filtering: {exc}")
            raise
        finally:
            if not current_uow:
                conn.commit()
                conn.close()

        return result

    def _create(self, model_obj):
        """ Add a new record to the sqlalchemy database"""
        conn = self._get_session()

        try:
            conn.add(model_obj)
        except DatabaseError as exc:
            logger.error(f"Error while creating: {exc}")
            raise
        finally:
            if not current_uow:
                conn.commit()
                conn.close()

        return model_obj

    def _update(self, model_obj):
        """ Update a record in the sqlalchemy database"""
        conn = self._get_session()
        db_item = None

        # Fetch the record from database
        try:
            identifier = getattr(
                model_obj, self.entity_cls.meta_.id_field.attribute_name
            )
            db_item = conn.query(self.model_cls).get(
                identifier
            )  # This will raise exception if object was not found
        except DatabaseError as exc:
            logger.error(f"Database Record not found: {exc}")
            raise

        if db_item is None:
            conn.rollback()
            conn.close()
            raise ObjectNotFoundError(
                f"`{self.entity_cls.__name__}` object with identifier {identifier} "
                f"does not exist."
            )

        # Sync DB Record with current changes. When the session is committed, changes are automatically synced
        try:
            for attribute in self.entity_cls.meta_.attributes:
                if attribute != self.entity_cls.meta_.id_field.attribute_name and getattr(
                    model_obj, attribute
                ) != getattr(
                    db_item, attribute
                ):
                    setattr(db_item, attribute, getattr(model_obj, attribute))
        except DatabaseError as exc:
            logger.error(f"Error while updating: {exc}")
            raise
        finally:
            if not current_uow:
                conn.commit()
                conn.close()

        return model_obj

    def _update_all(self, criteria: Q, *args, **kwargs):
        """ Update all objects satisfying the criteria """
        conn = self._get_session()
        qs = conn.query(self.model_cls).filter(self._build_filters(criteria))
        try:
            values = {}
            if args:
                values = args[
                    0
                ]  # `args[0]` is required because `*args` is sent as a tuple
            values.update(kwargs)
            updated_count = qs.update(values)
        except DatabaseError as exc:
            logger.error(f"Error while updating all: {exc}")
            raise
        finally:
            if not current_uow:
                conn.commit()
                conn.close()

        return updated_count

    def _delete(self, model_obj):
        """ Delete the entity record in the dictionary """
        conn = self._get_session()
        db_item = None

        # Fetch the record from database
        try:
            identifier = getattr(
                model_obj, self.entity_cls.meta_.id_field.attribute_name
            )
            db_item = conn.query(self.model_cls).get(
                identifier
            )  # This will raise exception if object was not found
        except DatabaseError as exc:
            logger.error(f"Database Record not found: {exc}")
            raise

        if db_item is None:
            conn.rollback()
            conn.close()
            raise ObjectNotFoundError(
                f"`{self.entity_cls.__name__}` object with identifier {identifier} "
                f"does not exist."
            )

        try:
            conn.delete(db_item)
        except DatabaseError as exc:
            logger.error(f"Error while deleting: {exc}")
            raise
        finally:
            if not current_uow:
                conn.commit()
                conn.close()

        return model_obj

    def _delete_all(self, criteria: Q = None):
        """ Delete a record from the sqlalchemy database"""
        conn = self._get_session()

        del_count = 0
        if criteria:
            qs = conn.query(self.model_cls).filter(self._build_filters(criteria))
        else:
            qs = conn.query(self.model_cls)

        try:
            del_count = qs.delete()
        except DatabaseError as exc:
            logger.error(f"Error while deleting all: {exc}")
            raise
        finally:
            if not current_uow:
                conn.commit()
                conn.close()

        return del_count

    def _raw(self, query: Any, data: Any = None):
        """Run a raw query on the repository and return entity objects"""
        assert isinstance(query, str)

        conn = self._get_session()
        try:
            results = conn.execute(query)

            entity_items = []
            for item in results:
                entity = self.model_cls.to_entity(item)
                entity.state_.mark_retrieved()
                entity_items.append(entity)

            result = ResultSet(
                offset=0,
                limit=len(entity_items),
                total=len(entity_items),
                items=entity_items,
            )
        except DatabaseError as exc:
            logger.error(f"Error while running raw query: {exc}")
            raise
        finally:
            if not current_uow:
                conn.commit()
                conn.close()

        return result


class SAProvider(BaseProvider):
    """Provider Implementation class for SQLAlchemy"""

    def __init__(self, *args, **kwargs):
        """Initialize and maintain Engine"""
        # Since SQLAlchemyProvider can cater to multiple databases, it is important
        #   that we know which database we are dealing with, to run database-specific
        #   statements like `PRAGMA` for SQLite.
        if "DATABASE" not in args[2]:
            logger.error(f"Missing `DATABASE` information in conn_info: {args[2]}")
            raise ConfigurationError("Missing `DATABASE` attribute in Connection info")

        super().__init__(*args, **kwargs)

        kwargs = self._get_database_specific_engine_args()

        self._engine = create_engine(make_url(self.conn_info["DATABASE_URI"]), **kwargs)

        if self.conn_info["DATABASE"] == Database.POSTGRESQL.value:
            # Nest database tables under a schema, so that we have complete control
            #   on creating/dropping db structures. We cannot control structures in the
            #   the default `public` schema.
            #
            # Use `SCHEMA` value if specified as part of the conn info. Otherwise, construct
            #   and use default schema name as `DB`_schema.
            schema = (
                self.conn_info["SCHEMA"] if "SCHEMA" in self.conn_info else "public"
            )

            self._metadata = MetaData(bind=self._engine, schema=schema)
        else:
            self._metadata = MetaData(bind=self._engine)

        # A temporary cache of already constructed model classes
        self._model_classes = {}

    def _get_database_specific_engine_args(self):
        """ Supplies additional database-specific arguments to SQLAlchemy Engine.

        Return: a dictionary with database-specific SQLAlchemy Engine arguments.
        """
        if self.conn_info["DATABASE"] == Database.POSTGRESQL.value:
            return {"isolation_level": "AUTOCOMMIT"}

        return {}

    def _get_database_specific_session_args(self):
        """ Set Database specific session parameters.

        Depending on the database in use, this method supplies
        additional arguments while constructing sessions.

        Return: a dictionary with additional arguments and values.
        """
        if self.conn_info["DATABASE"] == Database.POSTGRESQL.value:
            return {"autocommit": True, "autoflush": False}

        return {}

    def get_session(self):
        """Establish a session to the Database"""
        # Create the session
        kwargs = self._get_database_specific_session_args()
        session_factory = orm.sessionmaker(
            bind=self._engine, expire_on_commit=False, **kwargs
        )
        session_cls = orm.scoped_session(session_factory)

        return session_cls

    def _execute_database_specific_connection_statements(self, conn):
        """ Execute connection statements depending on the database in use.

        Each database has a unique set of commands and associated format to control
        connection-related parameters. Since we use SQLAlchemy, statements should
        be run dynamically based on the database in use.

        Arguments:
        * conn: An active connection object to the database

        Return: None
        """
        if self.conn_info["DATABASE"] == Database.SQLITE.value:
            conn.execute("PRAGMA case_sensitive_like = ON;")

        return conn

    def get_connection(self, session_cls=None):
        """ Create the connection to the Database instance"""
        # If this connection has to be created within an existing session,
        #   ``session_cls`` will be provided as an argument.
        #   Otherwise, fetch a new ``session_cls`` from ``get_session()``
        if session_cls is None:
            session_cls = self.get_session()

        conn = session_cls()
        conn = self._execute_database_specific_connection_statements(conn)

        return conn

    def _data_reset(self):
        conn = self._engine.connect()

        transaction = conn.begin()

        if self.conn_info["DATABASE"] == Database.SQLITE.value:
            conn.execute("PRAGMA foreign_keys = OFF;")

        for table in self._metadata.sorted_tables:
            conn.execute(table.delete())

        if self.conn_info["DATABASE"] == Database.SQLITE.value:
            conn.execute("PRAGMA foreign_keys = ON;")

        transaction.commit()

        # Discard any active Unit of Work
        if current_uow and current_uow.in_progress:
            current_uow.rollback()

    def decorate_model_class(self, entity_cls, model_cls):
        schema_name = derive_schema_name(model_cls)

        # Return the model class if it was already seen/decorated
        if schema_name in self._model_classes:
            return self._model_classes[schema_name]

        # If `model_cls` is already subclassed from SqlAlchemyModel,
        #   this method call is a no-op
        if issubclass(model_cls, SqlalchemyModel):
            return model_cls
        else:
            custom_attrs = {
                key: value
                for (key, value) in vars(model_cls).items()
                if key not in ["Meta", "__module__", "__doc__", "__weakref__"]
            }

            # Protean
            from protean.core.model import ModelMeta

            meta_ = ModelMeta()
            meta_.entity_cls = entity_cls

            custom_attrs.update({"meta_": meta_, "metadata": self._metadata})
            # FIXME Ensure the custom model attributes are constructed properly
            decorated_model_cls = type(
                model_cls.__name__, (SqlalchemyModel, model_cls), custom_attrs
            )

            # Memoize the constructed model class
            self._model_classes[schema_name] = decorated_model_cls

            return decorated_model_cls

    def construct_model_class(self, entity_cls):
        """Return a fully-baked Model class for a given Entity class"""
        model_cls = None

        # Return the model class if it was already seen/decorated
        if entity_cls.meta_.schema_name in self._model_classes:
            model_cls = self._model_classes[entity_cls.meta_.schema_name]
        else:
            # Protean
            from protean.core.model import ModelMeta

            meta_ = ModelMeta()
            meta_.entity_cls = entity_cls

            attrs = {
                "meta_": meta_,
                "metadata": self._metadata,
            }
            # FIXME Ensure the custom model attributes are constructed properly
            model_cls = type(entity_cls.__name__ + "Model", (SqlalchemyModel,), attrs)

            # Memoize the constructed model class
            self._model_classes[entity_cls.meta_.schema_name] = model_cls

        # Set Entity Class as a class level attribute for the Model, to be able to reference later.
        return model_cls

    def get_dao(self, entity_cls, model_cls):
        """ Return a DAO object configured with a live connection"""
        return SADAO(self.domain, self, entity_cls, model_cls)

    def raw(self, query: Any, data: Any = None):
        """Run raw query on Provider"""
        if data is None:
            data = {}
        assert isinstance(query, str)
        assert isinstance(data, (dict, None))

        return self.get_connection().execute(query, data)


operators = {
    "exact": "__eq__",
    "iexact": "ilike",
    "contains": "contains",
    "icontains": "ilike",
    "startswith": "startswith",
    "endswith": "endswith",
    "gt": "__gt__",
    "gte": "__ge__",
    "lt": "__lt__",
    "lte": "__le__",
    "in": "in_",
    "any": "any",
    "overlap": "overlap",
}


class DefaultLookup(BaseLookup):
    """Base class with default implementation of expression construction"""

    def __init__(self, source, target, model_cls):
        """Source is LHS and Target is RHS of a comparsion"""
        self.model_cls = model_cls
        super().__init__(source, target)

    def process_source(self):
        """Return source with transformations, if any"""
        source_col = getattr(self.model_cls, self.source)
        return source_col

    def process_target(self):
        """Return target with transformations, if any"""
        return self.target

    def as_expression(self):
        lookup_func = getattr(self.process_source(), operators[self.lookup_name])
        return lookup_func(self.process_target())


@SAProvider.register_lookup
class Exact(DefaultLookup):
    """Exact Match Query"""

    lookup_name = "exact"


@SAProvider.register_lookup
class IExact(DefaultLookup):
    """Exact Case-Insensitive Match Query"""

    lookup_name = "iexact"


@SAProvider.register_lookup
class Contains(DefaultLookup):
    """Exact Contains Query"""

    lookup_name = "contains"


@SAProvider.register_lookup
class IContains(DefaultLookup):
    """Exact Case-Insensitive Contains Query"""

    lookup_name = "icontains"

    def process_target(self):
        """Return target in lowercase"""
        assert isinstance(self.target, str)
        return f"%{super().process_target()}%"


@SAProvider.register_lookup
class Startswith(DefaultLookup):
    """Exact Contains Query"""

    lookup_name = "startswith"


@SAProvider.register_lookup
class Endswith(DefaultLookup):
    """Exact Contains Query"""

    lookup_name = "endswith"


@SAProvider.register_lookup
class GreaterThan(DefaultLookup):
    """Greater than Query"""

    lookup_name = "gt"


@SAProvider.register_lookup
class GreaterThanOrEqual(DefaultLookup):
    """Greater than or Equal Query"""

    lookup_name = "gte"


@SAProvider.register_lookup
class LessThan(DefaultLookup):
    """Less than Query"""

    lookup_name = "lt"


@SAProvider.register_lookup
class LessThanOrEqual(DefaultLookup):
    """Less than or Equal Query"""

    lookup_name = "lte"


@SAProvider.register_lookup
class In(DefaultLookup):
    """In Query"""

    lookup_name = "in"

    def process_target(self):
        """Ensure target is a list or tuple"""
        assert isinstance(self.target, (list, tuple))
        return super().process_target()


@SAProvider.register_lookup
class Any(DefaultLookup):
    """Any Query"""

    lookup_name = "any"


@SAProvider.register_lookup
class Overlap(DefaultLookup):
    """Overlap Query"""

    lookup_name = "overlap"



# Protean
import pytest

from protean.core.exceptions import InvalidOperationError

# Local/Relative Imports
from .elements import CustomBaseContainer, CustomContainer


class TestContainerInitialization:
    def test_that_base_container_class_cannot_be_instantiated(self):
        with pytest.raises(TypeError):
            CustomBaseContainer()

    def test_that_a_concrete_custom_container_can_be_instantiated(self):
        custom = CustomContainer(foo="a", bar="b")
        assert custom is not None


class TestContainerProperties:
    def test_two_containers_with_equal_values_are_considered_equal(self):
        custom1 = CustomContainer(foo="a", bar="b")
        custom2 = CustomContainer(foo="a", bar="b")

        assert custom1 == custom2

    @pytest.mark.xfail
    def test_that_container_objects_are_immutable(self):
        custom = CustomContainer(foo="a", bar="b")
        with pytest.raises(InvalidOperationError):
            custom.foo = "c"

    def test_output_to_dict(self):
        custom = CustomContainer(foo="a", bar="b")
        assert custom.to_dict() == {"foo": "a", "bar": "b"}

    @pytest.mark.xfail
    def test_that_only_valid_attributes_can_be_assigned(self):
        custom = CustomContainer(foo="a", bar="b")
        with pytest.raises(AttributeError):
            custom.foo = "bar"



# Protean
import pytest

from mock import patch
from protean.adapters.broker.inline import InlineBroker
from protean.core.domain_event import BaseDomainEvent
from protean.globals import current_domain
from protean.infra.event_log import EventLog, EventLogRepository
from protean.utils import fully_qualified_name

# Local/Relative Imports
from .elements import Person, PersonAdded, PersonCommand, PersonService


class TestDomainEventInitialization:
    def test_that_base_domain_event_class_cannot_be_instantiated(self):
        with pytest.raises(TypeError):
            BaseDomainEvent()

    def test_that_domain_event_can_be_instantiated(self):
        service = PersonAdded()
        assert service is not None


class TestDomainEventRegistration:
    def test_that_domain_event_can_be_registered_with_domain(self, test_domain):
        test_domain.register(PersonAdded)

        assert fully_qualified_name(PersonAdded) in test_domain.registry.domain_events

    def test_that_domain_event_can_be_registered_via_annotations(self, test_domain):
        @test_domain.domain_event(aggregate_cls=Person)
        class AnnotatedDomainEvent:
            def special_method(self):
                pass

        assert (
            fully_qualified_name(AnnotatedDomainEvent)
            in test_domain.registry.domain_events
        )


class TestDomainEventTriggering:
    @patch.object(InlineBroker, "send_message")
    def test_that_domain_event_is_raised_in_aggregate_command_method(self, mock):
        newcomer = Person.add_newcomer(
            {"first_name": "John", "last_name": "Doe", "age": 21}
        )
        mock.assert_called_once_with(PersonAdded(person=newcomer))

    def test_that_domain_event_is_persisted(self, test_domain):
        test_domain.register(Person)
        test_domain.register(EventLog)
        test_domain.register(EventLogRepository)

        command = PersonCommand(first_name="John", last_name="Doe", age=21)
        person = PersonService.add(command)

        event_repo = current_domain.repository_for(EventLog)
        event = event_repo.get_most_recent_event_by_type(kind=PersonAdded)

        assert event is not None
        assert event.kind == "PersonAdded"
        assert event.payload["person"] == person.to_dict()

    def test_that_all_domain_events_are_retrievable(self, test_domain):
        test_domain.register(Person)
        test_domain.register(EventLog)
        test_domain.register(EventLogRepository)

        command = PersonCommand(first_name="John", last_name="Doe", age=21)
        person = PersonService.add(command)

        event_repo = current_domain.repository_for(EventLog)
        events = event_repo.get_all_events_of_type(kind=PersonAdded)

        assert events is not None
        assert isinstance(events, list)
        assert len(events) == 1
        assert events[0].kind == "PersonAdded"
        assert events[0].payload["person"] == person.to_dict()



# Protean
import pytest

from mock import patch
from protean.core.subscriber import BaseSubscriber
from protean.utils import fully_qualified_name

# Local/Relative Imports
from .elements import NotifySSOSubscriber, Person, PersonAdded


class TestSubscriberInitialization:
    def test_that_base_subscriber_class_cannot_be_instantiated(self):
        with pytest.raises(TypeError):
            BaseSubscriber()

    def test_that_subscriber_can_be_instantiated(self, test_domain):
        service = NotifySSOSubscriber(test_domain, PersonAdded())
        assert service is not None


class TestApplicationServiceRegistration:
    def test_that_domain_event_can_be_registered_with_domain(self, test_domain):
        test_domain.register(NotifySSOSubscriber)

        assert (
            fully_qualified_name(NotifySSOSubscriber)
            in test_domain.registry.subscribers
        )

    def test_that_domain_event_can_be_registered_via_annotations(self, test_domain):
        @test_domain.subscriber(domain_event=PersonAdded)
        class AnnotatedSubscriber:
            def special_method(self):
                pass

        assert (
            fully_qualified_name(AnnotatedSubscriber)
            in test_domain.registry.subscribers
        )


class TestDomainEventNotification:
    @patch.object(NotifySSOSubscriber, "notify")
    def test_that_domain_event_is_received_from_aggregate_command_method(
        self, mock, test_domain
    ):
        test_domain.register(NotifySSOSubscriber)

        newcomer = Person.add_newcomer(
            {"first_name": "John", "last_name": "Doe", "age": 21}
        )
        mock.assert_called_once_with(PersonAdded(person=newcomer).to_dict())



# Protean
from protean.core.application_service import BaseApplicationService


class DummyApplicationService(BaseApplicationService):
    def do_application_process(self):
        print("Performing application process...")



"""Base class for Providers"""

# Standard Library Imports
from abc import ABCMeta, abstractmethod
from typing import Any

# Protean
from protean.utils.query import RegisterLookupMixin


class BaseProvider(RegisterLookupMixin, metaclass=ABCMeta):
    """Provider Implementation for each database that acts as a gateway to configure the database,
    retrieve connections and perform commits
    """

    def __init__(self, name, domain, conn_info: dict):
        """Initialize Provider with Connection/Adapter details"""
        self.name = name
        self.domain = domain
        self.conn_info = conn_info

    def _extract_lookup(self, key):
        """Extract lookup method based on key name format"""
        parts = key.split("__")
        # 'exact' is the default lookup if there was no explicit comparison op in `key`
        #   Assume there is only one `__` in the key.
        #   FIXME Change for child attribute query support
        op = "exact" if len(parts) == 1 else parts[1]

        # Construct and assign the lookup class as a filter criteria
        return parts[0], self.get_lookup(op)

    @abstractmethod
    def get_session(self):
        """Establish a new session with the database.

        Typically the session factory should be created once per application. Which is then
        held on to and passed to different transactions.

        In Protean's case, the session scope and the transaction scope match. Which means that a
        new session is created when a transaction needs to be initiated (at the beginning of
        request handling, for example) and terminated (after committing or rolling back) at the end
        of the process. The session will be used as a component in Unit of Work Pattern, to handle
        transactions reliably.

        Sessions are made available to requests as part of a Context Manager.
        """

    @abstractmethod
    def get_connection(self):
        """Get the connection object for the repository"""

    @abstractmethod
    def get_dao(self, entity_cls, model_cls):
        """Return a DAO object configured with a live connection"""

    @abstractmethod
    def decorate_model_class(self, entity_cls, model_cls):
        """Return decorated Model Class for custom-defined models"""

    @abstractmethod
    def construct_model_class(self, entity_cls):
        """Return dynamically constructed Model Class"""

    @abstractmethod
    def raw(self, query: Any, data: Any = None):
        """Run raw query directly on the database

        Query should be executed immediately on the database as a separate unit of work
            (in a different transaction context). The results should be returned as returned by
            the database without any intervention. It is left to the consumer to interpret and
            organize the results correctly.
        """



# Protean
import pytest

from protean.adapters.repository.sqlalchemy import SqlalchemyModel

# Local/Relative Imports
from .elements import ComplexUser, Email, Person, Provider, Receiver


@pytest.mark.sqlite
class TestModel:
    @pytest.fixture(autouse=True)
    def register_person_aggregate(self, test_domain):
        test_domain.register(Person)

    def test_that_model_class_is_created_automatically(self, test_domain):
        model_cls = test_domain.get_model(Person)
        assert issubclass(model_cls, SqlalchemyModel)
        assert model_cls.__name__ == "PersonModel"

    def test_conversation_from_entity_to_model(self, test_domain):
        model_cls = test_domain.get_model(Person)
        person = Person(first_name="John", last_name="Doe")
        person_model_obj = model_cls.from_entity(person)

        assert person_model_obj is not None
        assert isinstance(person_model_obj, SqlalchemyModel)
        assert person_model_obj.age == 21
        assert person_model_obj.first_name == "John"
        assert person_model_obj.last_name == "Doe"

    def test_conversation_from_model_to_entity(self, test_domain):
        model_cls = test_domain.get_model(Person)
        person = Person(first_name="John", last_name="Doe")
        person_model_obj = model_cls.from_entity(person)

        person_copy = model_cls.to_entity(person_model_obj)
        assert person_copy is not None

    def test_dynamically_constructed_model_attributes(self, test_domain):
        # Protean
        from sqlalchemy import String

        model_cls = test_domain.get_model(Person)

        assert model_cls.__name__ == "PersonModel"
        assert type(model_cls.first_name.type) is String


@pytest.mark.sqlite
class TestModelWithVO:
    @pytest.fixture(autouse=True)
    def register_complex_user_aggregate(self, test_domain):
        test_domain.register(ComplexUser)

    def test_that_model_class_is_created_automatically(self, test_domain):
        model_cls = test_domain.get_model(ComplexUser)
        assert issubclass(model_cls, SqlalchemyModel)
        assert model_cls.__name__ == "ComplexUserModel"

    def test_conversation_from_entity_to_model(self, test_domain):
        model_cls = test_domain.get_model(ComplexUser)

        user1 = ComplexUser(email_address="john.doe@gmail.com", password="d4e5r6")
        user2 = ComplexUser(
            email=Email(address="john.doe@gmail.com"), password="d4e5r6"
        )

        user1_model_obj = model_cls.from_entity(user1)
        user2_model_obj = model_cls.from_entity(user2)

        assert user1_model_obj is not None
        assert isinstance(user1_model_obj, SqlalchemyModel)
        assert user1_model_obj.email_address == "john.doe@gmail.com"
        assert user1_model_obj.password == "d4e5r6"

        assert user2_model_obj is not None
        assert isinstance(user2_model_obj, SqlalchemyModel)
        assert user2_model_obj.email_address == "john.doe@gmail.com"
        assert user2_model_obj.password == "d4e5r6"

        # Model's content should reflect only the attributes, not declared_fields
        assert hasattr(user1_model_obj, "email") is False
        assert hasattr(user2_model_obj, "email") is False

    def test_conversation_from_model_to_entity(self, test_domain):
        model_cls = test_domain.get_model(ComplexUser)
        user1 = ComplexUser(email_address="john.doe@gmail.com", password="d4e5r6")
        user1_model_obj = model_cls.from_entity(user1)

        user_copy = model_cls.to_entity(user1_model_obj)
        assert user_copy is not None


@pytest.mark.sqlite
class TestCustomModel:
    def test_that_custom_model_can_be_associated_with_entity(self, test_domain):
        model_cls = test_domain.get_model(Provider)
        assert model_cls.__name__ == "ProviderCustomModel"

    def test_that_model_can_be_registered_with_domain_annotation(self, test_domain):
        # Protean
        from sqlalchemy import Column, Text

        test_domain.register(Receiver)

        @test_domain.model(entity_cls=Receiver)
        class ReceiverInlineModel:
            name = Column(Text)

        test_domain.get_dao(Receiver)

        provider = test_domain.get_provider("default")
        provider._metadata.create_all()

        model_cls = test_domain.get_model(Receiver)
        assert model_cls.__name__ == "ReceiverInlineModel"

        assert type(model_cls.name.type) is Text



# Protean
import pytest

from protean.globals import current_domain

# Local/Relative Imports
from .child_entities import Comment, Post, PostMeta


class TestHasOnePersistence:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Post)
        test_domain.register(PostMeta)
        test_domain.register(Comment)

    @pytest.fixture(autouse=True)
    def persist_post(self, test_domain, register_elements):
        post = test_domain.get_dao(Post).create(
            title="Test Post", slug="test-post", content="Do Re Mi Fa"
        )
        return post

    @pytest.fixture
    def persisted_post(self, test_domain):
        return test_domain.get_dao(Post).find_by(title="Test Post")

    def test_that_has_one_entity_can_be_added(self, persisted_post):
        post_repo = current_domain.repository_for(Post)

        meta = PostMeta(likes=1)
        persisted_post.post_meta = meta

        post_repo.add(persisted_post)

        refreshed_post = post_repo.get(persisted_post.id)
        assert refreshed_post is not None
        assert refreshed_post.post_meta is not None
        assert isinstance(refreshed_post.post_meta, PostMeta)
        assert refreshed_post.post_meta == meta

    def test_that_adding_another_has_one_entity_replaces_existing_child(
        self, persisted_post
    ):
        post_repo = current_domain.repository_for(Post)

        meta1 = PostMeta(likes=1)
        meta2 = PostMeta(likes=2)
        persisted_post.post_meta = meta1

        post_repo.add(persisted_post)

        post_to_alter = post_repo.get(persisted_post.id)
        post_to_alter.post_meta = meta2

        post_repo.add(post_to_alter)

        refreshed_post = post_repo.get(persisted_post.id)

        assert refreshed_post is not None
        assert refreshed_post.post_meta is not None
        assert isinstance(refreshed_post.post_meta, PostMeta)
        assert refreshed_post.post_meta == meta2

    def test_that_a_has_one_entity_can_be_removed(self, persisted_post):
        post_repo = current_domain.repository_for(Post)

        meta = PostMeta(likes=1)
        persisted_post.post_meta = meta

        post_repo.add(persisted_post)

        post_to_alter = post_repo.get(persisted_post.id)
        post_to_alter.post_meta = None

        post_repo.add(post_to_alter)

        refreshed_post = post_repo.get(persisted_post.id)
        assert refreshed_post is not None
        assert refreshed_post.post_meta is None


class TestHasManyPersistence:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Post)
        test_domain.register(PostMeta)
        test_domain.register(Comment)

    @pytest.fixture
    def persisted_post(self, test_domain):
        post = test_domain.get_dao(Post).create(
            title="Test Post", slug="test-post", content="Do Re Mi Fa"
        )
        return post

    def test_that_a_has_many_entity_can_be_added(self, persisted_post):
        post_repo = current_domain.repository_for(Post)

        comment = Comment(content="So La Ti Do")
        persisted_post.comments.add(comment)

        post_repo.add(persisted_post)

        refreshed_post = post_repo.get(persisted_post.id)
        assert refreshed_post is not None
        assert refreshed_post.comments is not None
        assert comment.id in [comment.id for comment in refreshed_post.comments]

    def test_that_multiple_has_many_entities_can_be_added(self, persisted_post):
        post_repo = current_domain.repository_for(Post)

        comment1 = Comment(content="So La Ti Do")
        comment2 = Comment(content="Do Re Mi Fa")
        persisted_post.comments.add(comment1)
        persisted_post.comments.add(comment2)

        post_repo.add(persisted_post)

        refreshed_post = post_repo.get(persisted_post.id)
        assert refreshed_post is not None
        assert refreshed_post.comments is not None
        assert len(refreshed_post.comments) == 2
        assert all(
            comment in [comment for comment in refreshed_post.comments]
            for comment in [comment1, comment2]
        )

    def test_that_a_has_many_entity_can_be_removed(self, persisted_post):
        post_repo = current_domain.repository_for(Post)

        comment = Comment(content="So La Ti Do")
        persisted_post.comments.add(comment)

        post_repo.add(persisted_post)

        post_to_alter = post_repo.get(persisted_post.id)
        post_to_alter.comments.remove(comment)

        post_repo.add(post_to_alter)

        refreshed_post = post_repo.get(persisted_post.id)
        assert refreshed_post is not None
        assert refreshed_post.comments is not None
        assert len(refreshed_post.comments) == 0

    def test_that_a_has_many_entity_can_be_removed_from_among_many(
        self, persisted_post
    ):
        post_repo = current_domain.repository_for(Post)

        comment1 = Comment(content="So La Ti Do")
        comment2 = Comment(content="Do Re Mi Fa")
        persisted_post.comments.add(comment1)
        persisted_post.comments.add(comment2)

        post_repo.add(persisted_post)

        post_to_alter = post_repo.get(persisted_post.id)
        post_to_alter.comments.remove(comment1)

        post_repo.add(post_to_alter)

        refreshed_post = post_repo.get(persisted_post.id)
        assert refreshed_post is not None
        assert refreshed_post.comments is not None
        assert len(refreshed_post.comments) == 1
        assert comment2.id in [comment.id for comment in refreshed_post.comments]



# Standard Library Imports
from abc import abstractmethod

# Protean
from protean.globals import current_domain
from protean.utils import (
    DomainObjects,
    convert_str_values_to_list,
    derive_element_class,
)


class BaseEmailProvider:
    """
    Base class for email backend implementations.

    Concrete implementations must overwrite `send_messages()`.
    ```
    """

    def __init__(self, name, domain, conn_info, fail_silently=False, **kwargs):
        self.name = name
        self.domain = domain
        self.conn_info = conn_info
        self.fail_silently = fail_silently

    @abstractmethod
    def send_email(self, email_messages):
        """
        Send one or more EmailMessage objects and return the number of email
        messages sent.
        """
        raise NotImplementedError(
            "Concrete implementations of BaseEmailBackend "
            "must override send_messages() method"
        )


class _EmailMetaclass(type):
    """
    This base metaclass processes the class declaration and constructs a meta object that can
    be used to introspect the Email class later. Specifically, it sets up a `meta_` attribute on
    the Email to an instance of Meta, either the default of one that is defined in the
    Email class.

    `meta_` is setup with these attributes:
        * `provider`: The email provider that this email message is associated with
    """

    def __new__(mcs, name, bases, attrs, **kwargs):
        """Initialize Email MetaClass and load attributes"""

        # Ensure initialization is only performed for subclasses of Email
        # (excluding Email class itself).
        parents = [b for b in bases if isinstance(b, _EmailMetaclass)]
        if not parents:
            return super().__new__(mcs, name, bases, attrs)

        # Remove `abstract` in base classes if defined
        for base in bases:
            if hasattr(base, "Meta") and hasattr(base.Meta, "abstract"):
                delattr(base.Meta, "abstract")

        new_class = super().__new__(mcs, name, bases, attrs, **kwargs)

        # Gather `Meta` class/object if defined
        attr_meta = attrs.pop("Meta", None)
        meta = attr_meta or getattr(new_class, "Meta", None)
        setattr(new_class, "meta_", EmailMeta(name, meta))

        return new_class


class EmailMeta:
    """ Metadata info for the Email.

    Options:
    - ``provider``: The Email provider that this message is associated with
    """

    def __init__(self, entity_name, meta):
        self.provider = getattr(meta, "provider", "default")


class BaseEmail(metaclass=_EmailMetaclass):
    """Base Email class that should implemented by all Domain Email Messages.

    This is also a marker class that is referenced when emails are registered
    with the domain.
    """

    element_type = DomainObjects.EMAIL

    def __new__(cls, *args, **kwargs):
        if cls is BaseEmail:
            raise TypeError("BaseEmail cannot be instantiated")
        return super().__new__(cls)

    def __init__(
        self,
        subject="",
        data="",
        from_email=None,
        to=None,
        bcc=None,
        cc=None,
        reply_to=None,
        template_id="",
        **kwargs
    ):
        """
        Initialize a single email message (which can be sent to multiple
        recipients).
        """
        self.to = convert_str_values_to_list(to)
        self.cc = convert_str_values_to_list(cc)
        self.bcc = convert_str_values_to_list(bcc)

        provider = current_domain.get_email_provider(self.meta_.provider)
        self.provider = provider
        self.from_email = from_email or provider.conn_info["DEFAULT_FROM_EMAIL"]

        self.reply_to = (
            convert_str_values_to_list(reply_to) if reply_to else self.from_email
        )

        self.subject = subject
        self.template_id = template_id
        self.data = data
        self.kwargs = kwargs

    @property
    def mime_message(self):
        """ Convert the message to a mime compliant email string """
        return "\n".join([self.from_email, str(self.to), self.subject[:25]])

    def __repr__(self):
        """ Convert the message to a mime compliant email string """
        return "\n".join([self.from_email, str(self.to), self.subject[:25]])

    @property
    def recipients(self):
        """
        Return a list of all recipients of the email (includes direct
        addressees as well as Cc and Bcc entries).
        """
        return [email for email in (self.to + self.cc + self.bcc) if email]


def email_factory(element_cls, **kwargs):
    return derive_element_class(element_cls, BaseEmail)



"""Aggregate Functionality and Classes"""
# Standard Library Imports
import logging

# Protean
from protean.core.entity import BaseEntity
from protean.core.exceptions import NotSupportedError
from protean.utils import DomainObjects, derive_element_class

logger = logging.getLogger("protean.domain.aggregate")


class BaseAggregate(BaseEntity):
    """The Base class for Protean-Compliant Domain Aggregates.

    Provides helper methods to custom define aggregate attributes, and query attribute names
    during runtime.

    Basic Usage::

        @domain.aggregate
        class Dog:
            id = field.Integer(identifier=True)
            name = field.String(required=True, max_length=50)
            age = field.Integer(default=5)
            owner = field.String(required=True, max_length=15)

    (or)

        class Dog(BaseAggregate):
            id = field.Integer(identifier=True)
            name = field.String(required=True, max_length=50)
            age = field.Integer(default=5)
            owner = field.String(required=True, max_length=15)

        domain.register_element(Dog)

    During persistence, the model associated with this entity is retrieved dynamically from
            the repository factory. Model is usually initialized with a live DB connection.
    """

    element_type = DomainObjects.AGGREGATE

    def __new__(cls, *args, **kwargs):
        if cls is BaseAggregate:
            raise TypeError("BaseAggregate cannot be instantiated")
        return super().__new__(cls)


def aggregate_factory(element_cls, **kwargs):
    element_cls = derive_element_class(element_cls, BaseAggregate)

    if element_cls.meta_.abstract is True:
        raise NotSupportedError(
            f"{element_cls.__name__} class has been marked abstract"
            f" and cannot be instantiated"
        )

    element_cls.meta_.provider = (
        kwargs.pop("provider", None)
        or (hasattr(element_cls, "meta_") and element_cls.meta_.provider)
        or "default"
    )
    element_cls.meta_.model = (
        kwargs.pop("model", None)
        or (hasattr(element_cls, "meta_") and element_cls.meta_.model)
        or None
    )

    return element_cls



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

# Standard Library Imports
import os
import sys

from os.path import abspath, dirname, exists, join

if __name__ == "__main__":
    base_path = dirname(dirname(abspath(__file__)))
    print("Project path: {0}".format(base_path))
    env_path = join(base_path, ".tox", "bootstrap")
    if sys.platform == "win32":
        bin_path = join(env_path, "Scripts")
    else:
        bin_path = join(env_path, "bin")
    if not exists(env_path):
        import subprocess

        print("Making bootstrap env in: {0} ...".format(env_path))
        try:
            subprocess.check_call(["virtualenv", env_path])
        except subprocess.CalledProcessError:
            subprocess.check_call([sys.executable, "-m", "virtualenv", env_path])
        print("Installing `jinja2` into bootstrap environment...")
        subprocess.check_call([join(bin_path, "pip"), "install", "jinja2"])
    activate = join(bin_path, "activate_this.py")
    # noinspection PyCompatibility
    exec(
        compile(open(activate, "rb").read(), activate, "exec"), dict(__file__=activate)
    )

    import jinja2

    import subprocess

    jinja = jinja2.Environment(
        loader=jinja2.FileSystemLoader(join(base_path, "ci", "templates")),
        trim_blocks=True,
        lstrip_blocks=True,
        keep_trailing_newline=True,
    )

    tox_environments = [
        line.strip()
        # WARNING: 'tox' must be installed globally or in the project's virtualenv
        for line in subprocess.check_output(
            ["tox", "--listenvs"], universal_newlines=True
        ).splitlines()
    ]
    tox_environments = [
        line
        for line in tox_environments
        if line not in ["clean", "report", "docs", "check"]
    ]

    for name in os.listdir(join("ci", "templates")):
        with open(join(base_path, name), "w") as fh:
            fh.write(jinja.get_template(name).render(tox_environments=tox_environments))
        print("Wrote {}".format(name))
    print("DONE.")



# Protean
import pytest

from protean.adapters.repository import elasticsearch as repo


@pytest.mark.elasticsearch
class TestLookup:
    def test_exact_lookup(self, test_domain):
        lookup = repo.Exact("first_name", "John")

        assert lookup.as_expression().to_dict() == {"term": {"first_name": "John"}}

    def test_in_lookup(self, test_domain):
        lookup = repo.In("first_name", ["John"])

        assert lookup.as_expression().to_dict() == {"terms": {"first_name": ["John"]}}

    def test_gt_lookup(self, test_domain):
        lookup = repo.GreaterThan("age", 6)

        assert lookup.as_expression().to_dict() == {"range": {"age": {"gt": 6}}}

    def test_gte_lookup(self, test_domain):
        lookup = repo.GreaterThanOrEqual("age", 6)

        assert lookup.as_expression().to_dict() == {"range": {"age": {"gte": 6}}}

    def test_lt_lookup(self, test_domain):
        lookup = repo.LessThan("age", 6)

        assert lookup.as_expression().to_dict() == {"range": {"age": {"lt": 6}}}

    def test_lte_lookup(self, test_domain):
        lookup = repo.LessThanOrEqual("age", 6)

        assert lookup.as_expression().to_dict() == {"range": {"age": {"lte": 6}}}

    def test_contains_lookup(self, test_domain):
        lookup = repo.Contains("first_name", "John")

        assert lookup.as_expression().to_dict() == {
            "wildcard": {"first_name": {"value": "*John*"}}
        }

    def test_startswith_lookup(self, test_domain):
        lookup = repo.Startswith("first_name", "John")

        assert lookup.as_expression().to_dict() == {
            "wildcard": {"first_name": {"value": "John*"}}
        }

    def test_endswith_lookup(self, test_domain):
        lookup = repo.Endswith("first_name", "John")

        assert lookup.as_expression().to_dict() == {
            "wildcard": {"first_name": {"value": "*John"}}
        }



# Protean
from protean.core.aggregate import BaseAggregate
from protean.core.domain_event import BaseDomainEvent
from protean.core.field.basic import Auto, Integer, String
from protean.core.subscriber import BaseSubscriber
from protean.globals import current_domain


class Person(BaseAggregate):
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50, required=True)
    age = Integer(default=21)

    @classmethod
    def add_newcomer(cls, person_dict):
        """Factory method to add a new Person to the system"""
        newcomer = Person(
            first_name=person_dict["first_name"],
            last_name=person_dict["last_name"],
            age=person_dict["age"],
        )

        # Publish Event via the domain
        current_domain.publish(
            PersonAdded(
                id=newcomer.id,
                first_name=newcomer.first_name,
                last_name=newcomer.last_name,
                age=newcomer.age,
            )
        )

        return newcomer


class PersonAdded(BaseDomainEvent):
    id = Auto(identifier=True)
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50, required=True)
    age = Integer(default=21)


class NotifySSOSubscriber(BaseSubscriber):
    """Subscriber that notifies an external SSO system
    that a new person was added into the system
    """

    class Meta:
        domain_event = PersonAdded

    def notify(self, domain_event_dict):
        print("Received Domain Event: ", domain_event_dict)
        print("Domain Event class: ", self.meta_.domain_event)

        print("Current domain: ", current_domain.domain_name)



# cSpell: disable

# Protean
from protean.utils import Database, IdentityStrategy, IdentityType

DEBUG = True
TESTING = True
ENV = "development"

# A secret key for this particular Protean installation. Used in secret-key
# hashing algorithms.
SECRET_KEY = "BuyDV45%6R%hdNqDvex@6nB7@yscjQta"

# Database Configuration
DATABASES = {
    "default": {
        "PROVIDER": "protean.adapters.repository.sqlalchemy.SAProvider",
        "DATABASE": Database.SQLITE.value,
        "DATABASE_URI": "sqlite:///test.db",
    },
}

IDENTITY_STRATEGY = IdentityStrategy.UUID
IDENTITY_TYPE = IdentityType.STRING

# Messaging Mediums
BROKERS = {
    "default": {"PROVIDER": "protean.adapters.InlineBroker"},
}



# Standard Library Imports
import os

# Protean
import pytest


@pytest.fixture
def test_domain_with_string_identity():
    # Protean
    from protean.domain import Domain

    domain = Domain("Test")

    # Construct relative path to config file
    current_path = os.path.abspath(os.path.dirname(__file__))
    config_path = os.path.join(current_path, "./config_string.py")

    if os.path.exists(config_path):
        domain.config.from_pyfile(config_path)

    with domain.domain_context():
        yield domain


@pytest.fixture
def test_domain_with_int_identity():
    # Protean
    from protean.domain import Domain

    domain = Domain("Test")

    # Construct relative path to config file
    current_path = os.path.abspath(os.path.dirname(__file__))
    config_path = os.path.join(current_path, "./config_int.py")

    if os.path.exists(config_path):
        domain.config.from_pyfile(config_path)

    with domain.domain_context():
        yield domain


@pytest.fixture
def test_domain_with_uuid_identity():
    # Protean
    from protean.domain import Domain

    domain = Domain("Test")

    # Construct relative path to config file
    current_path = os.path.abspath(os.path.dirname(__file__))
    config_path = os.path.join(current_path, "./config_uuid.py")

    if os.path.exists(config_path):
        domain.config.from_pyfile(config_path)

    with domain.domain_context():
        yield domain


@pytest.fixture(autouse=True)
def run_around_tests(test_domain):

    yield

    if test_domain.providers.has_provider("default"):
        test_domain.get_provider("default")._data_reset()



# Protean
import pytest

from protean.core.exceptions import InvalidOperationError
from protean.core.unit_of_work import UnitOfWork


class TestUnitOfWorkInitialization:
    def test_uow_can_be_initiated_with_context_manager(self, test_domain):
        with UnitOfWork() as uow:
            assert uow is not None
            assert uow.in_progress is True

    def test_uow_can_be_initiated_explicitly(self, test_domain):
        uow = UnitOfWork()
        assert uow is not None
        assert uow.in_progress is False

        uow.start()
        assert uow.in_progress is True

    def test_that_uow_throws_exception_on_commit_or_rollback_without_being_started(
        self, test_domain
    ):
        uow = UnitOfWork()

        with pytest.raises(InvalidOperationError):
            uow.commit()

        with pytest.raises(InvalidOperationError):
            uow.rollback()

    def test_that_uow_can_be_started_manually(self, test_domain):
        uow = UnitOfWork()

        uow.start()
        uow.commit()  # `commit` should not raise exception

        uow.start()
        uow.rollback()  # `rollback` should not raise exception



# Standard Library Imports
from datetime import datetime

# Protean
import pytest

from protean.core.aggregate import BaseAggregate
from protean.core.field.basic import DateTime, Dict, String
from protean.globals import current_domain
from sqlalchemy import types as sa_types


class Event(BaseAggregate):
    name = String(max_length=255)
    created_at = DateTime(default=datetime.utcnow())
    payload = Dict()


@pytest.mark.postgresql
def test_json_data_type_association(test_domain):
    test_domain.register(Event)

    model_cls = test_domain.get_model(Event)
    type(model_cls.payload.property.columns[0].type) == sa_types.JSON


@pytest.mark.postgresql
def test_basic_dict_data_type_operations(test_domain):
    test_domain.register(Event)

    model_cls = test_domain.get_model(Event)

    event = Event(
        name="UserCreated", payload={"email": "john.doe@gmail.com", "password": "*****"}
    )
    event_model_obj = model_cls.from_entity(event)

    event_copy = model_cls.to_entity(event_model_obj)
    assert event_copy is not None
    assert event_copy.payload == {"email": "john.doe@gmail.com", "password": "*****"}


@pytest.mark.postgresql
def test_json_with_array_data(test_domain):
    test_domain.register(Event)

    model_cls = test_domain.get_model(Event)

    event = Event(
        name="UserCreated",
        payload=[
            {"email": "john.doe@gmail.com", "password": "*****"},
            {"email": "john.doe1234@gmail.com", "password": "*****"},
        ],
    )
    event_model_obj = model_cls.from_entity(event)

    event_copy = model_cls.to_entity(event_model_obj)
    assert event_copy is not None
    assert event_copy.payload == [
        {"email": "john.doe@gmail.com", "password": "*****"},
        {"email": "john.doe1234@gmail.com", "password": "*****"},
    ]


@pytest.mark.postgresql
def test_persistence_of_json_with_array_data(test_domain):
    test_domain.register(Event)

    event = Event(
        name="UserCreated",
        payload=[
            {"email": "john.doe@gmail.com", "password": "*****"},
            {"email": "john.doe1234@gmail.com", "password": "*****"},
        ],
    )
    current_domain.get_dao(Event).save(event)

    refreshed_event = current_domain.get_dao(Event).get(event.id)
    assert refreshed_event is not None
    assert refreshed_event.payload == [
        {"email": "john.doe@gmail.com", "password": "*****"},
        {"email": "john.doe1234@gmail.com", "password": "*****"},
    ]



# Protean
import pytest

from protean.core.unit_of_work import UnitOfWork

# Local/Relative Imports
from .aggregate_elements import Comment, Post, PostMeta, PostRepository


class TestUnitOfWorkRegistration:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Post)
        test_domain.register(PostMeta)
        test_domain.register(Comment)

        test_domain.register(PostRepository, aggregate_cls=Post)

        yield

    @pytest.fixture
    def persisted_post(self, test_domain):
        post = test_domain.get_dao(Post).create(
            title="Test Post", slug="test-post", content="Do Re Mi Fa"
        )
        return post

    def test_that_an_entity_can_be_added_within_uow(self, test_domain, persisted_post):
        repo = test_domain.repository_for(Post)

        with UnitOfWork():
            comment = Comment(content="So La Ti Do")
            persisted_post.comments.add(comment)

            repo = test_domain.repository_for(Post)
            repo.add(persisted_post)

            # FIXME Refactor `outside_uow` to be a global thread variable
            # post_dao = test_domain.get_dao(Post)
            # assert len(post_dao.outside_uow().get(persisted_post.id).comments) == 0

        post = repo.get(persisted_post.id)
        assert len(post.comments) == 1
        assert post.comments[0].content == "So La Ti Do"

    def test_that_an_entity_can_be_updated_within_uow(
        self, test_domain, persisted_post
    ):
        comment = Comment(content="So La Ti Do")
        persisted_post.comments.add(comment)

        repo = test_domain.repository_for(Post)
        repo.add(persisted_post)

        post = repo.get(persisted_post.id)
        assert post.comments[0].content == "So La Ti Do"

        with UnitOfWork():
            comment = persisted_post.comments[0]
            comment.content = "Pa Da Ni Sa"
            persisted_post.comments.add(comment)

            repo = test_domain.repository_for(Post)
            repo.add(persisted_post)

            # FIXME Refactor `outside_uow` to be a global thread variable
            # assert comment.id in uow.changes_to_be_committed['default']['UPDATED']

        post = repo.get(persisted_post.id)
        assert post.comments[0].content == "Pa Da Ni Sa"

    def test_that_an_entity_can_be_removed_within_uow(
        self, test_domain, persisted_post
    ):
        comment = Comment(content="So La Ti Do")
        persisted_post.comments.add(comment)

        repo = test_domain.repository_for(Post)
        repo.add(persisted_post)

        with UnitOfWork():
            comment = persisted_post.comments[0]
            persisted_post.comments.remove(comment)

            repo = test_domain.repository_for(Post)
            repo.add(persisted_post)

            # FIXME Refactor `outside_uow` to be a global thread variable
            # assert comment.id in uow.changes_to_be_committed['default']['REMOVED']

        post = repo.get(persisted_post.id)
        assert len(post.comments) == 0



""" Utilities for converting names between different cases"""

# Standard Library Imports
import re


def camelize(string, uppercase_first_letter=True):
    """
    Convert strings to CamelCase.

    Examples::

        >>> camelize("device_type")
        "DeviceType"
        >>> camelize("device_type", False)
        "deviceType"

    :func:`camelize` can be though as a inverse of :func:`underscore`, although
    there are some cases where that does not hold::

        >>> camelize(underscore("IOError"))
        "IoError"

    :param uppercase_first_letter: if set to `True` :func:`camelize` converts
        strings to UpperCamelCase. If set to `False` :func:`camelize` produces
        lowerCamelCase. Defaults to `True`.
    """
    if uppercase_first_letter:
        return re.sub(r"(?:^|_)(.)", lambda m: m.group(1).upper(), string)
    else:
        return string[0].lower() + camelize(string)[1:]


def dasherize(word):
    """Replace underscores with dashes in the string.

    Example::

        >>> dasherize("lower_case")
        "lower-case"

    """
    return word.replace("_", "-")


def humanize(word):
    """
    Capitalize the first word and turn underscores into spaces and strip a
    trailing ``"_id"``, if any. Like :func:`titleize`, this is meant for
    creating pretty output.

    Examples::

        >>> humanize("employee_salary")
        "Employee salary"
        >>> humanize("author_id")
        "Author"

    """
    word = re.sub(r"_id$", "", word)
    word = word.replace("_", " ")
    word = re.sub(r"(?i)([a-z\d]*)", lambda m: m.group(1).lower(), word)
    word = re.sub(r"^\w", lambda m: m.group(0).upper(), word)
    return word


def titleize(word):
    """
    Capitalize all the words and replace some characters in the string to
    create a nicer looking title. :func:`titleize` is meant for creating pretty
    output.

    Examples::

      >>> titleize("man from the boondocks")
      "Man From The Boondocks"
      >>> titleize("x-men: the last stand")
      "X Men: The Last Stand"
      >>> titleize("TheManWithoutAPast")
      "The Man Without A Past"
      >>> titleize("raiders_of_the_lost_ark")
      "Raiders Of The Lost Ark"

    """
    return re.sub(
        r"\b('?[a-z])",
        lambda match: match.group(1).capitalize(),
        humanize(underscore(word)),
    )


def underscore(word):
    """
    Make an underscored, lowercase form from the expression in the string.

    Example::

        >>> underscore("DeviceType")
        "device_type"

    As a rule of thumb you can think of :func:`underscore` as the inverse of
    :func:`camelize`, though there are cases where that does not hold::

        >>> camelize(underscore("IOError"))
        "IoError"

    """
    word = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", word)
    word = re.sub(r"([a-z\d])([A-Z])", r"\1_\2", word)
    word = word.replace("-", "_")
    return word.lower()



# Protean
import pytest

# Local/Relative Imports
from .aggregate_elements import Comment, Post, PostMeta


class TestAggregatesWithEntities:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Post)
        test_domain.register(PostMeta)
        test_domain.register(Comment)

    @pytest.fixture
    def persisted_post(self, test_domain):
        post = test_domain.get_dao(Post).create(
            title="Test Post", slug="test-post", content="Do Re Mi Fa"
        )
        return post

    def test_that_an_entity_can_be_added(self, persisted_post):
        comment = Comment(content="So La Ti Do")
        persisted_post.comments.add(comment)

        assert comment in persisted_post.comments

    def test_that_the_parent_is_associated_with_child_once_added_to_parent(
        self, persisted_post
    ):
        comment = Comment(content="So La Ti Do")
        persisted_post.comments.add(comment)

        assert comment.post_id is not None

    def test_that_adding_an_existing_entity_does_not_create_duplicates(
        self, persisted_post
    ):
        comment = Comment(content="So La Ti Do")
        persisted_post.comments.add(comment)

        assert comment in persisted_post.comments
        assert len(persisted_post.comments) == 1

        # Add the child object again
        persisted_post.comments.add(comment)

        assert comment in persisted_post.comments
        assert len(persisted_post.comments) == 1

    def test_that_an_entity_can_be_removed(self, persisted_post):
        comment = Comment(content="So La Ti Do")
        persisted_post.comments.add(comment)

        assert comment in persisted_post.comments

        persisted_post.comments.remove(comment)
        assert comment not in persisted_post.comments

    def test_that_one_entity_amongst_many_can_be_removed(self, persisted_post):
        comment1 = Comment(content="So La Ti Do")
        comment2 = Comment(content="Sa Re Ga Ma")
        persisted_post.comments.add(comment1)
        persisted_post.comments.add(comment2)

        assert comment1 in persisted_post.comments
        assert comment2 in persisted_post.comments
        assert len(persisted_post.comments) == 2

        persisted_post.comments.remove(comment1)

        assert len(persisted_post.comments) == 1
        assert comment1 not in persisted_post.comments
        assert comment2 in persisted_post.comments
        assert comment2.id == persisted_post.comments[0].id



# Protean
import pytest

from protean.core.exceptions import ValidationError

# Local/Relative Imports
from .elements import Account, Author, Post, Profile


class TestReferenceFieldAssociation:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Account)
        test_domain.register(Author)
        test_domain.register(Post)
        test_domain.register(Profile)

    def test_initialization_of_an_entity_containing_reference_field(self, test_domain):
        account = Account(email="john.doe@gmail.com", password="a1b2c3")
        author = Author(first_name="John", last_name="Doe", account=account)

        assert all(
            key in author.__dict__
            for key in ["first_name", "last_name", "account_email"]
        )
        assert author.account.email == account.email

    def test_successful_save_of_entity_with_reference_field(self, test_domain):
        account = Account(email="john.doe@gmail.com", password="a1b2c3")
        author = Author(first_name="John", last_name="Doe", account=account)

        assert not author.state_.is_persisted
        test_domain.get_dao(Author).save(author)

        assert author.state_.is_persisted

    @pytest.mark.skip
    def test_that_assignment_fails_when_an_invalid_entity_type_is_assigned_to_a_relation(
        self, test_domain
    ):
        class DummyAccount(Account):
            pass

        account = DummyAccount(email="john.doe@gmail.com", password="a1b2c3")
        with pytest.raises(ValidationError):
            Author(first_name="John", last_name="Doe", account=account)

    def test_the_presence_of_shadow_attribute_backing_the_association(
        self, test_domain
    ):
        account = Account(email="john.doe@gmail.com", password="a1b2c3")
        test_domain.get_dao(Account).save(account)
        author = Author(first_name="John", last_name="Doe", account=account)
        test_domain.get_dao(Author).save(author)

        assert all(key in author.__dict__ for key in ["account", "account_email"])
        assert author.account.email == account.email
        assert author.account_email == account.email

    def test_save_after_value_assignment_during_initialization(self, test_domain):
        account = Account(email="john.doe@gmail.com", password="a1b2c3")
        test_domain.get_dao(Account).save(account)
        author = Author(first_name="John", last_name="Doe", account=account)
        test_domain.get_dao(Author).save(author)

        assert all(key in author.__dict__ for key in ["account", "account_email"])
        assert author.account.email == account.email
        assert author.account_email == account.email

    def test_save_after_explicit_reference_value_assignment(self, test_domain):
        account = Account(email="john.doe@gmail.com", password="a1b2c3")
        test_domain.get_dao(Account).save(account)

        author = Author(first_name="John", last_name="Doe")
        assert (
            any(key in author.__dict__ for key in ["account", "account_email"]) is False
        )

        # Explicitly assign value to reference field
        author.account = account
        test_domain.get_dao(Author).save(author)

        assert all(key in author.__dict__ for key in ["account", "account_email"])
        assert author.account_email == account.email

    def test_fetch_after_save_and_ensure_reference_is_not_auto_loaded(
        self, test_domain
    ):
        account = Account(email="john.doe@gmail.com", password="a1b2c3")
        test_domain.get_dao(Account).save(account)
        author = Author(first_name="John", last_name="Doe", account=account)
        test_domain.get_dao(Author).save(author)

        author = test_domain.get_dao(Author).get(author.id)
        # Reference attribute is not loaded automatically
        assert "account" not in author.__dict__
        assert author.account_email == account.email

        # Accessing attribute shows it up in __dict__
        getattr(author, "account")
        assert "account" in author.__dict__

    def test_value_assignment_via_shadow_attribute_during_initialization(
        self, test_domain
    ):
        account = Account(email="john.doe@gmail.com", password="a1b2c3")
        test_domain.get_dao(Account).save(account)

        author = Author(first_name="John", last_name="Doe", account_email=account.email)
        assert "account_email" in author.__dict__
        assert "account" not in author.__dict__

        test_domain.get_dao(Author).save(author)

        assert author.account.email == account.email
        assert author.account_email == account.email

        # This `keys` test coming after checking `account_email` values is important
        #   because accessing the property will load the key into __dict__
        assert all(key in author.__dict__ for key in ["account", "account_email"])

    def test_value_assignment_via_shadow_attribute_post_initialization(
        self, test_domain
    ):
        account = Account(email="john.doe@gmail.com", password="a1b2c3")
        test_domain.get_dao(Account).save(account)

        author = Author(first_name="John", last_name="Doe")
        author.account_email = account.email
        assert "account" not in author.__dict__

        test_domain.get_dao(Author).save(author)

        assert author.account.email == account.email
        assert author.account_email == account.email

        # This `keys` test coming after checking `account_email` values is important
        #   because accessing the property will load the key into __dict__
        assert "account" in author.__dict__

    def test_that_resetting_reference_field_resets_shadow_attribute_too(
        self, test_domain
    ):
        account = Account(email="john.doe@gmail.com", password="a1b2c3")
        test_domain.get_dao(Account).save(account)
        author = Author(first_name="John", last_name="Doe", account=account)

        assert author.account.email == account.email
        assert author.account_email == account.email

        author.account = None
        assert (
            any(key in author.__dict__ for key in ["account", "account_email"]) is False
        )
        assert author.account is None
        assert "account_email" not in author.__dict__

    def test_that_setting_shadow_attribute_to_none_resets_reference_field_too(
        self, test_domain
    ):
        account = Account(email="john.doe@gmail.com", password="a1b2c3")
        test_domain.get_dao(Account).save(account)
        author = Author(first_name="John", last_name="Doe", account=account)

        assert author.account.email == account.email
        assert author.account_email == account.email

        assert "account_email" in author.meta_.attributes
        author.account_email = None

        assert (
            any(key in author.__dict__ for key in ["account", "account_email"]) is False
        )
        assert author.account is None
        assert author.account_email is None
        assert "account_email" not in author.__dict__

    def test_that_resetting_the_reference_field_resets_shadow_attribute_too(
        self, test_domain
    ):
        account = Account(email="john.doe@gmail.com", password="a1b2c3")
        test_domain.get_dao(Account).save(account)
        author = Author(first_name="John", last_name="Doe", account=account)

        del author.account
        assert (
            any(key in author.__dict__ for key in ["account", "account_email"]) is False
        )
        assert author.account is None
        assert author.account_email is None
        assert "account_email" not in author.__dict__

    def test_that_resetting_the_shadow_attribute_resets_reference_field_too(
        self, test_domain
    ):
        account = Account(email="john.doe@gmail.com", password="a1b2c3")
        test_domain.get_dao(Account).save(account)
        author = Author(first_name="John", last_name="Doe", account=account)

        del author.account_email
        assert (
            any(key in author.__dict__ for key in ["account", "account_email"]) is False
        )
        assert author.account is None
        assert author.account_email is None
        assert "account_email" not in author.__dict__

    def test_successful_save_with_an_entity_linked_by_via(self, test_domain):
        account = Account(
            email="john.doe@gmail.com", password="a1b2c3", username="johndoe"
        )
        test_domain.get_dao(Account).save(account)
        profile = Profile(about_me="Lorem Ipsum", account=account)
        test_domain.get_dao(Profile).save(profile)

        assert all(key in profile.__dict__ for key in ["account", "account_username"])
        assert hasattr(profile, "account_username")
        assert profile.account_username == account.username

    def test_successful_save_with_an_entity_linked_by_via_and_assigned_by_shadow_attribute(
        self, test_domain
    ):
        account = Account(
            email="john.doe@gmail.com", password="a1b2c3", username="johndoe"
        )
        test_domain.get_dao(Account).save(account)
        profile = Profile(about_me="Lorem Ipsum")
        profile.account_username = account.username
        test_domain.get_dao(Profile).save(profile)

        assert hasattr(profile, "account_username")
        assert profile.account.email == account.email
        assert profile.account_username == account.username

    def test_that_subsequent_accesses_after_first_retrieval_do_not_fetch_record_again(
        self, test_domain
    ):
        account = Account(
            email="john.doe@gmail.com", password="a1b2c3", username="johndoe"
        )
        test_domain.get_dao(Account).save(account)
        author = Author(first_name="John", last_name="Doe", account_email=account.email)

        for _ in range(3):
            getattr(author, "account")



# Protean
from protean.utils import fully_qualified_name


class TestVORegistration:
    def test_auto_register_value_object_with_annotation(self, test_domain):
        # Protean
        from protean.core.field.basic import String

        @test_domain.aggregate
        class Foo:
            foo = String()

        @test_domain.value_object(aggregate_cls=Foo)
        class Bar:
            bar = String()

        assert fully_qualified_name(Bar) in test_domain.registry.value_objects
        assert Bar.meta_.aggregate_cls == Foo



# Protean
from protean.core.aggregate import BaseAggregate
from protean.core.command import BaseCommand
from protean.core.command_handler import BaseCommandHandler
from protean.core.field.basic import Integer, String


class Person(BaseAggregate):
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50, required=True)
    age = Integer(default=21)

    @classmethod
    def add_newcomer(cls, person_dict):
        """Factory method to add a new Person to the system"""
        newcomer = Person(
            first_name=person_dict["first_name"],
            last_name=person_dict["last_name"],
            age=person_dict["age"],
        )

        return newcomer


class AddPersonCommand(BaseCommand):
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50, required=True)
    age = Integer(default=21)


class AddNewPersonCommandHandler(BaseCommandHandler):
    """CommandHandler that adds a new person into the system"""

    class Meta:
        command_cls = AddPersonCommand

    def notify(self, command):
        print("Received command: ", command)



"""Module containing repository implementation for Elasticsearch"""
# Standard Library Imports
import logging

from typing import Any
from uuid import UUID

# Protean
import elasticsearch_dsl

from elasticsearch import Elasticsearch
from elasticsearch.exceptions import NotFoundError
from elasticsearch_dsl import Document, Index, Search, query
from protean.core.exceptions import ObjectNotFoundError
from protean.core.field.association import Reference
from protean.globals import current_domain
from protean.port.dao import BaseDAO, BaseLookup, ResultSet
from protean.port.provider import BaseProvider
from protean.utils import Database, IdentityStrategy, IdentityType
from protean.utils.query import Q

logger = logging.getLogger("protean.repository")

operators = {
    "exact": "__eq__",
    "iexact": "ilike",
    "contains": "contains",
    "icontains": "ilike",
    "startswith": "startswith",
    "endswith": "endswith",
    "gt": "__gt__",
    "gte": "__ge__",
    "lt": "__lt__",
    "lte": "__le__",
    "in": "in_",
    "overlap": "overlap",
    "any": "any",
}


def derive_schema_name(model_cls):
    if hasattr(model_cls.meta_, "schema_name"):
        return model_cls.meta_.schema_name
    else:
        return model_cls.meta_.entity_cls.meta_.schema_name


class ElasticsearchModel(Document):
    """A model for the Elasticsearch index"""

    @classmethod
    def from_entity(cls, entity) -> "ElasticsearchModel":
        """Convert the entity to a Elasticsearch record """
        item_dict = {}
        for attribute_obj in cls.meta_.entity_cls.meta_.attributes.values():
            if isinstance(attribute_obj, Reference):
                item_dict[
                    attribute_obj.relation.attribute_name
                ] = attribute_obj.relation.value
            else:
                item_dict[attribute_obj.attribute_name] = getattr(
                    entity, attribute_obj.attribute_name
                )

        model_obj = cls(**item_dict)

        if "id" in item_dict:
            model_obj.meta.id = model_obj.id
            del model_obj._d_["id"]  # pylint: disable=W0212

        return model_obj

    @classmethod
    def to_entity(cls, item: "ElasticsearchModel"):
        """Convert the elasticsearch document to an entity """
        item_dict = {}

        # Convert the values in ES Model as a dictionary
        values = item.to_dict()
        for field_name in cls.meta_.entity_cls.meta_.attributes:
            item_dict[field_name] = values.get(field_name, None)

        identifier = None
        if (
            current_domain.config["IDENTITY_STRATEGY"] == IdentityStrategy.UUID
            and current_domain.config["IDENTITY_TYPE"] == IdentityType.UUID
            and isinstance(item.meta.id, str)
        ):
            identifier = UUID(item.meta.id)
        else:
            identifier = item.meta.id

        item_dict["id"] = identifier
        entity_obj = cls.meta_.entity_cls(item_dict)

        return entity_obj


class ESSession:
    """A Session wrapper for Elasticsearch Database.

    Elasticsearch does not support Transactions or Sessions, so this class is
    essential a no-op, and acts as a passthrough for all transactions.
    """

    def __init__(self, provider, new_connection=False):
        self._provider = provider

    def add(self, element):
        dao = self._provider.get_dao(element.__class__)
        dao.create(element.to_dict())

    def delete(self, element):
        dao = self._provider.get_dao(element.__class__)
        dao.delete(element)

    def commit(self):
        pass

    def rollback(self):
        pass

    def close(self):
        pass


class ElasticsearchDAO(BaseDAO):
    def __repr__(self) -> str:
        return f"ElasticsearchDAO <{self.entity_cls.__name__}>"

    def get(self, identifier: Any):
        """Retrieve a specific Record from the Repository by its `identifier`.

        This method internally uses the `filter` method to fetch records.

        Returns exactly one record that matches the identifier.

        Throws `ObjectNotFoundError` if no record was found for the identifier.

        Throws `TooManyObjectsError` if multiple records were found for the identifier.

        :param identifier: id of the record to be fetched from the data store.
        """
        logger.debug(
            f"Lookup `{self.entity_cls.__name__}` object with identifier {identifier}"
        )

        conn = self._get_session()
        result = None

        try:
            result = self.model_cls.get(
                id=identifier, using=conn, index=self.entity_cls.meta_.schema_name
            )
        except NotFoundError:
            logger.error(f"Record {identifier} was not found")
            raise ObjectNotFoundError(
                f"`{self.entity_cls.__name__}` object with identifier {identifier} "
                f"does not exist."
            )
        except Exception as error:
            logger.error(f"Unknown error occurred when fetching {identifier}: {error}")
            raise (
                f"`{self.entity_cls.__name__}` object with identifier {identifier} "
                f"does not exist."
            )

        entity = self.model_cls.to_entity(result)
        return entity

    def _build_filters(self, criteria: Q):
        """ Recursively Build the filters from the criteria object"""
        composed_query = query.Q()

        if criteria.connector == criteria.AND:
            for child in criteria.children:
                if isinstance(child, Q):
                    composed_query = composed_query & self._build_filters(child)
                else:
                    stripped_key, lookup_class = self.provider._extract_lookup(child[0])
                    lookup = lookup_class(stripped_key, child[1])
                    if criteria.negated:
                        composed_query = composed_query & ~lookup.as_expression()
                    else:
                        composed_query = composed_query & lookup.as_expression()
        else:
            for child in criteria.children:
                if isinstance(child, Q):
                    composed_query = composed_query | self._build_filters(child)
                else:
                    stripped_key, lookup_class = self.provider._extract_lookup(child[0])
                    lookup = lookup_class(stripped_key, child[1])
                    if criteria.negated:
                        composed_query = composed_query | ~lookup.as_expression()
                    else:
                        composed_query = composed_query | lookup.as_expression()

        return composed_query

    def _filter(
        self, criteria: Q, offset: int = 0, limit: int = 10, order_by: list = ()
    ) -> ResultSet:
        """
        Filter objects from the data store. Method must return a `ResultSet`
        object
        """
        conn = self._get_session()

        # Build the filters from the criteria
        q = elasticsearch_dsl.Q()
        if criteria.children:
            q = self._build_filters(criteria)

        s = Search(using=conn, index=self.entity_cls.meta_.schema_name).query(q)

        if order_by:
            s = s.sort(*order_by)

        s = s[offset : offset + limit]

        # Return the results
        try:
            response = s.execute()
            result = ResultSet(
                offset=offset,
                limit=limit,
                total=response.hits.total.value,
                items=response.hits,
            )
        except Exception as exc:
            logger.error(f"Error while filtering: {exc}")
            raise

        return result

    def _create(self, model_obj: Any):
        """Create a new model object from the entity"""
        conn = self._get_session()

        try:
            model_obj.save(
                refresh=True,
                index=model_obj.meta_.entity_cls.meta_.schema_name,
                using=conn,
            )
        except Exception as exc:
            logger.error(f"Error while creating: {exc}")
            raise

        return model_obj

    def _update(self, model_obj: Any):
        """Update a model object in the data store and return it"""
        conn = self._get_session()

        identifier = model_obj.meta.id

        # Fetch the record from database
        try:
            # Calling `get` will raise `NotFoundError` if record was not found
            self.model_cls.get(
                id=identifier, using=conn, index=self.entity_cls.meta_.schema_name
            )
        except NotFoundError as exc:
            logger.error(f"Database Record not found: {exc}")
            raise ObjectNotFoundError(
                f"`{self.entity_cls.__name__}` object with identifier {identifier} "
                f"does not exist."
            )

        try:
            model_obj.save(
                refresh=True,
                index=model_obj.meta_.entity_cls.meta_.schema_name,
                using=conn,
            )
        except Exception as exc:
            logger.error(f"Error while creating: {exc}")
            raise

        return model_obj

    def _update_all(self, criteria: Q, *args, **kwargs):
        """Updates object directly in the data store and returns update count"""
        raise NotImplementedError

    def _delete(self, model_obj):
        """Delete a Record from the Repository"""
        conn = self._get_session()

        try:
            model_obj.delete(
                index=model_obj.meta_.entity_cls.meta_.schema_name,
                using=conn,
                refresh=True,
            )
        except Exception as exc:
            logger.error(f"Error while creating: {exc}")
            raise

        return model_obj

    def _delete_all(self, criteria: Q = None):
        """Delete all records matching criteria from the Repository"""
        conn = self._get_session()

        # Build the filters from the criteria
        q = elasticsearch_dsl.Q()
        if criteria and criteria.children:
            q = self._build_filters(criteria)

        s = Search(using=conn, index=self.entity_cls.meta_.schema_name).query(q)

        # Return the results
        try:
            response = s.delete()

            # `Search.delete` does not refresh index, so we have to manually refresh
            index = Index(name=self.entity_cls.meta_.schema_name, using=conn)
            index.refresh()
        except Exception as exc:
            logger.error(f"Error while deleting records: {exc}")
            raise

        return response.deleted

    def _raw(self, query: Any, data: Any = None):
        """Run raw query on Data source.

        Running a raw query on the data store should always returns entity instance objects. If
        the results were not synthesizable back into entity objects, an exception should be thrown.
        """
        raise NotImplementedError


class ESProvider(BaseProvider):
    def __init__(self, name, domain, conn_info: dict):
        """Initialize Provider with Connection/Adapter details"""

        # In case of `ESProvider`, the `DATABASE` value will always be `ELASTICSEARCH`.
        conn_info["DATABASE"] = Database.ELASTICSEARCH.value
        super().__init__(name, domain, conn_info)

        # A temporary cache of already constructed model classes
        self._model_classes = {}

    def _extract_lookup(self, key):
        """Extract lookup method based on key name format"""
        parts = key.rsplit("__", 1)

        if len(parts) > 1 and parts[1] in operators:
            op = parts[1]
            attribute = parts[0]
        else:
            # 'exact' is the default lookup if there was no explicit comparison op in `key`
            op = "exact"
            attribute = key

        # Construct and assign the lookup class as a filter criteria
        return attribute, self.get_lookup(op)

    def get_session(self):
        """Establish a new session with the database.

        Typically the session factory should be created once per application. Which is then
        held on to and passed to different transactions.

        In Protean's case, the session scope and the transaction scope match. Which means that a
        new session is created when a transaction needs to be initiated (at the beginning of
        request handling, for example) and terminated (after committing or rolling back) at the end
        of the process. The session will be used as a component in Unit of Work Pattern, to handle
        transactions reliably.

        Sessions are made available to requests as part of a Context Manager.
        """
        return ESSession(self)

    def get_connection(self):
        """Get the connection object for the repository"""
        return Elasticsearch(self.conn_info["DATABASE_URI"]["hosts"])

    def get_dao(self, entity_cls, model_cls):
        """Return a DAO object configured with a live connection"""
        return ElasticsearchDAO(self.domain, self, entity_cls, model_cls)

    def decorate_model_class(self, entity_cls, model_cls):
        schema_name = derive_schema_name(model_cls)

        # Return the model class if it was already seen/decorated
        if schema_name in self._model_classes:
            return self._model_classes[schema_name]

        # If `model_cls` is already subclassed from SqlAlchemyModel,
        #   this method call is a no-op
        if issubclass(model_cls, ElasticsearchModel):
            return model_cls
        else:
            custom_attrs = {
                key: value
                for (key, value) in vars(model_cls).items()
                if key not in ["Meta", "__module__", "__doc__", "__weakref__"]
            }

            # Protean
            from protean.core.model import ModelMeta

            meta_ = ModelMeta()
            meta_.entity_cls = entity_cls

            custom_attrs.update({"meta_": meta_})
            # FIXME Ensure the custom model attributes are constructed properly
            decorated_model_cls = type(
                model_cls.__name__, (ElasticsearchModel, model_cls), custom_attrs
            )

            # Memoize the constructed model class
            self._model_classes[schema_name] = decorated_model_cls

            return decorated_model_cls

    def construct_model_class(self, entity_cls):
        """Return a fully-baked Model class for a given Entity class"""
        model_cls = None

        # Return the model class if it was already seen/decorated
        if entity_cls.meta_.schema_name in self._model_classes:
            model_cls = self._model_classes[entity_cls.meta_.schema_name]
        else:
            # Protean
            from protean.core.model import ModelMeta

            meta_ = ModelMeta()
            meta_.entity_cls = entity_cls

            attrs = {
                "meta_": meta_,
            }
            # FIXME Ensure the custom model attributes are constructed properly
            model_cls = type(
                entity_cls.__name__ + "Model", (ElasticsearchModel,), attrs
            )

            # Memoize the constructed model class
            self._model_classes[entity_cls.meta_.schema_name] = model_cls

        # Set Entity Class as a class level attribute for the Model, to be able to reference later.
        return model_cls

    def raw(self, query: Any, data: Any = None):
        """Run raw query directly on the database

        Query should be executed immediately on the database as a separate unit of work
            (in a different transaction context). The results should be returned as returned by
            the database without any intervention. It is left to the consumer to interpret and
            organize the results correctly.
        """
        raise NotImplementedError

    def _data_reset(self):
        """Utility method to reset data in DB between tests"""
        conn = self.get_connection()

        for _, aggregate_record in current_domain.registry.aggregates.items():
            provider = current_domain.get_provider(aggregate_record.cls.meta_.provider)
            if provider.conn_info["DATABASE"] == Database.ELASTICSEARCH.value:
                conn.delete_by_query(
                    refresh=True,
                    index=aggregate_record.cls.meta_.schema_name,
                    body={"query": {"match_all": {}}},
                )


class DefaultLookup(BaseLookup):
    """Base class with default implementation of expression construction"""

    def process_target(self):
        """Return target with transformations, if any"""
        if isinstance(self.target, UUID):
            self.target = str(self.target)

        return self.target


@ESProvider.register_lookup
class Exact(DefaultLookup):
    """Exact Match Query"""

    lookup_name = "exact"

    def as_expression(self):
        return query.Q("term", **{self.process_source(): self.process_target()})


@ESProvider.register_lookup
class In(DefaultLookup):
    """In Match Query"""

    lookup_name = "in"

    def process_target(self):
        """Ensure target is a list or tuple"""
        assert isinstance(self.target, (list, tuple))
        return super().process_target()

    def as_expression(self):
        return query.Q("terms", **{self.process_source(): self.process_target()})


@ESProvider.register_lookup
class GreaterThan(DefaultLookup):
    """Greater than Query"""

    lookup_name = "gt"

    def as_expression(self):
        return query.Q(
            "range", **{self.process_source(): {"gt": self.process_target()}}
        )


@ESProvider.register_lookup
class GreaterThanOrEqual(DefaultLookup):
    """Greater than or Equal Query"""

    lookup_name = "gte"

    def as_expression(self):
        return query.Q(
            "range", **{self.process_source(): {"gte": self.process_target()}}
        )


@ESProvider.register_lookup
class LessThan(DefaultLookup):
    """Less than Query"""

    lookup_name = "lt"

    def as_expression(self):
        return query.Q(
            "range", **{self.process_source(): {"lt": self.process_target()}}
        )


@ESProvider.register_lookup
class LessThanOrEqual(DefaultLookup):
    """Less than or Equal Query"""

    lookup_name = "lte"

    def as_expression(self):
        return query.Q(
            "range", **{self.process_source(): {"lte": self.process_target()}}
        )


@ESProvider.register_lookup
class Contains(DefaultLookup):
    """Exact Contains Query"""

    lookup_name = "contains"

    def as_expression(self):
        return query.Q(
            "wildcard",
            **{self.process_source(): {"value": f"*{self.process_target()}*"}},
        )


@ESProvider.register_lookup
class Startswith(DefaultLookup):
    """Exact Contains Query"""

    lookup_name = "startswith"

    def as_expression(self):
        return query.Q(
            "wildcard",
            **{self.process_source(): {"value": f"{self.process_target()}*"}},
        )


@ESProvider.register_lookup
class Endswith(DefaultLookup):
    """Exact Contains Query"""

    lookup_name = "endswith"

    def as_expression(self):
        return query.Q(
            "wildcard",
            **{self.process_source(): {"value": f"*{self.process_target()}"}},
        )



# Standard Library Imports
from datetime import datetime

# Protean
import pytest

from protean.core.aggregate import BaseAggregate
from protean.core.exceptions import ValidationError
from protean.core.field.basic import (
    Auto,
    Boolean,
    Date,
    DateTime,
    Float,
    Identifier,
    Integer,
    List,
    String,
)
from sqlalchemy import types as sa_types


class ArrayUser(BaseAggregate):
    email = String(max_length=255, required=True, unique=True)
    roles = List()  # Defaulted to String Content Type


class IntegerArrayUser(BaseAggregate):
    email = String(max_length=255, required=True, unique=True)
    roles = List(content_type=Integer)


@pytest.mark.sqlite
def test_array_data_type_association(test_domain):
    test_domain.register(ArrayUser)

    model_cls = test_domain.get_model(ArrayUser)
    type(model_cls.roles.property.columns[0].type) == sa_types.ARRAY


@pytest.mark.sqlite
def test_basic_array_data_type_operations(test_domain):
    test_domain.register(ArrayUser)

    model_cls = test_domain.get_model(ArrayUser)

    user = ArrayUser(email="john.doe@gmail.com", roles=["ADMIN", "USER"])
    user_model_obj = model_cls.from_entity(user)

    user_copy = model_cls.to_entity(user_model_obj)
    assert user_copy is not None
    assert user_copy.roles == ["ADMIN", "USER"]


@pytest.mark.sqlite
def test_array_content_type_validation(test_domain):
    test_domain.register(ArrayUser)
    test_domain.register(IntegerArrayUser)

    for kwargs in [
        {"email": "john.doe@gmail.com", "roles": [1, 2]},
        {"email": "john.doe@gmail.com", "roles": ["1", 2]},
        {"email": "john.doe@gmail.com", "roles": [1.0, 2.0]},
        {"email": "john.doe@gmail.com", "roles": [datetime.utcnow()]},
    ]:
        with pytest.raises(ValidationError) as exception:
            ArrayUser(**kwargs)
        assert exception.value.messages == {"roles": ["Invalid value"]}

    model_cls = test_domain.get_model(IntegerArrayUser)
    user = IntegerArrayUser(email="john.doe@gmail.com", roles=[1, 2])
    user_model_obj = model_cls.from_entity(user)

    user_copy = model_cls.to_entity(user_model_obj)
    assert user_copy is not None
    assert user_copy.roles == [1, 2]

    for kwargs in [
        {"email": "john.doe@gmail.com", "roles": ["ADMIN", "USER"]},
        {"email": "john.doe@gmail.com", "roles": ["1", "2"]},
        {"email": "john.doe@gmail.com", "roles": [datetime.utcnow()]},
    ]:
        with pytest.raises(ValidationError) as exception:
            IntegerArrayUser(**kwargs)
        assert exception.value.messages == {"roles": ["Invalid value"]}


@pytest.mark.sqlite
def test_that_only_specific_primitive_types_are_allowed_as_content_types(test_domain):
    List(content_type=String)
    List(content_type=Identifier)
    List(content_type=Integer)
    List(content_type=Float)
    List(content_type=Boolean)
    List(content_type=Date)
    List(content_type=DateTime)

    with pytest.raises(ValidationError) as error:
        List(content_type=Auto)

    assert error.value.messages == {"content_type": ["Content type not supported"]}



# Standard Library Imports
import copy
import logging

from collections import defaultdict

# Protean
from protean.core.exceptions import InvalidDataError, NotSupportedError, ValidationError
from protean.core.field.basic import Field

logger = logging.getLogger("protean.domain")


class _ContainerMetaclass(type):
    """
    This base metaclass processes the class declaration and constructs a meta object that can
    be used to introspect the concrete Container class later.

    It also sets up a `meta_` attribute on the concrete class to an instance of Meta,
    either the default of one that is defined in the concrete class.
    """

    def __new__(mcs, name, bases, attrs, **kwargs):
        """Initialize Container MetaClass and load attributes"""

        # Ensure initialization is only performed for subclasses of Container
        # (excluding Container class itself).
        parents = [b for b in bases if isinstance(b, _ContainerMetaclass)]
        if not parents:
            return super().__new__(mcs, name, bases, attrs)

        # Remove `abstract` in base classes if defined
        for base in bases:
            if hasattr(base, "Meta") and hasattr(base.Meta, "abstract"):
                delattr(base.Meta, "abstract")

        new_class = super().__new__(mcs, name, bases, attrs, **kwargs)

        # Gather `Meta` class/object if defined
        attr_meta = attrs.pop("Meta", None)
        meta = attr_meta or getattr(new_class, "Meta", None)
        setattr(new_class, "meta_", ContainerMeta(name, meta))

        # Load declared fields
        new_class._load_fields(attrs)

        # Load declared fields from Base class, in case this Entity is subclassing another
        new_class._load_base_class_fields(bases, attrs)

        return new_class

    def _load_base_class_fields(new_class, bases, attrs):
        """If this class is subclassing another Container, add that Container's
        fields.  Note that we loop over the bases in *reverse*.
        This is necessary in order to maintain the correct order of fields.
        """
        for base in reversed(bases):
            if hasattr(base, "meta_") and hasattr(base.meta_, "declared_fields"):
                base_class_fields = {
                    field_name: field_obj
                    for (field_name, field_obj) in base.meta_.declared_fields.items()
                    if field_name not in attrs and not field_obj.identifier
                }
                new_class._load_fields(base_class_fields)

    def _load_fields(new_class, attrs):
        """Load field items into Class"""
        for attr_name, attr_obj in attrs.items():
            if isinstance(attr_obj, Field):
                setattr(new_class, attr_name, attr_obj)
                new_class.meta_.declared_fields[attr_name] = attr_obj


class ContainerMeta:
    """ Metadata info for the Container.

    Options:
    - ``abstract``: Indicates that this is an abstract entity (Ignores all other meta options)

    Also acts as a placeholder for generated entity fields like:

        :declared_fields: dict
            Any instances of `Field` included as attributes on either the class
            or on any of its superclasses will be include in this dictionary.
    """

    def __init__(self, entity_name, meta):
        self.abstract = getattr(meta, "abstract", None) or False

        # FIXME These attributes should be decided per domain element
        #   Was added here to be able to move Domain Event Factory into its own file
        self.aggregate_cls = getattr(meta, "aggregate_cls", None)

        # Initialize Options
        self.declared_fields = {}

    @property
    def mandatory_fields(self):
        """ Return the mandatory fields for this entity """
        return {
            field_name: field_obj
            for field_name, field_obj in self.attributes.items()
            if field_obj.required
        }

    @property
    def attributes(self):
        attributes_dict = {}
        for field_name, field_obj in self.declared_fields.items():
            attributes_dict[field_obj.get_attribute_name()] = field_obj

        return attributes_dict


class BaseContainer(metaclass=_ContainerMetaclass):
    """The Base class for Protean-Compliant Data Containers.

    Provides helper methods to custom define attributes, and find attribute names
    during runtime.
    """

    def __new__(cls, *args, **kwargs):
        if cls is BaseContainer:
            raise TypeError("BaseContainer cannot be instantiated")
        return super().__new__(cls)

    def __init__(self, *template, owner=None, raise_errors=True, **kwargs):
        """
        Initialise the container.

        During initialization, set value on fields if validation passes.

        This initialization technique supports keyword arguments as well as dictionaries. You
            can even use a template for initial data.
        """

        if self.meta_.abstract is True:
            raise NotSupportedError(
                f"{self.__class__.__name__} class has been marked abstract"
                f" and cannot be instantiated"
            )

        self.errors = defaultdict(list)
        self.raise_errors = raise_errors

        # Entity/Aggregate to which this Command is connected to
        self.owner = owner

        # Load the attributes based on the template
        loaded_fields = []
        for dictionary in template:
            if not isinstance(dictionary, dict):
                raise AssertionError(
                    f'Positional argument "{dictionary}" passed must be a dict.'
                    f"This argument serves as a template for loading common "
                    f"values.",
                )
            for field_name, val in dictionary.items():
                loaded_fields.append(field_name)
                setattr(self, field_name, val)

        # Now load against the keyword arguments
        for field_name, val in kwargs.items():
            loaded_fields.append(field_name)
            setattr(self, field_name, val)

        # Now load the remaining fields with a None value, which will fail
        # for required fields
        for field_name, field_obj in self.meta_.declared_fields.items():
            if field_name not in loaded_fields:
                setattr(self, field_name, None)

        self.defaults()

        # `clean()` will return a `defaultdict(list)` if errors are to be raised
        custom_errors = self.clean() or {}
        for field in custom_errors:
            self.errors[field].extend(custom_errors[field])

        # Raise any errors found during load
        if self.errors and self.raise_errors:
            logger.error(self.errors)
            raise ValidationError(self.errors)

    @classmethod
    def build(cls, **values):
        if values:
            assert all(
                attr in list(cls.meta_.declared_fields.keys()) for attr in values.keys()
            )

        return cls(**values)

    def defaults(self):
        """Placeholder method for defaults.
        To be overridden in concrete Containers, when an attribute's default depends on other attribute values.
        """

    def clean(self):
        """Placeholder method for validations.
        To be overridden in concrete Containers, when complex validations spanning multiple fields are required.
        """
        return defaultdict(list)

    def __eq__(self, other):
        """Equivalence check for commands is based only on data.

        Two Commands are considered equal if they have the same data.
        """
        if type(other) is not type(self):
            return False

        return self.to_dict() == other.to_dict()

    def __hash__(self):
        """Overrides the default implementation and bases hashing on values"""
        return hash(frozenset(self.to_dict().items()))

    def __repr__(self):
        """Friendly repr for Command"""
        return "<%s: %s>" % (self.__class__.__name__, self)

    def __str__(self):
        return "%s object (%s)" % (
            self.__class__.__name__,
            "{}".format(self.to_dict()),
        )

    def __bool__(self):
        """ Return this object's truthiness to be `False`,
        if all its attributes evaluate to truthiness `False`
        """
        return any(
            bool(getattr(self, field_name, None))
            for field_name in self.meta_.attributes
        )

    def __setattr__(self, name, value):
        if name in self.meta_.declared_fields or name in [
            "raise_errors",
            "errors",
            "owner",
        ]:
            super().__setattr__(name, value)
        else:
            raise InvalidDataError({name: ["is invalid"]})

    def to_dict(self):
        """ Return data as a dictionary """
        return {
            field_name: getattr(self, field_name, None)
            for field_name in self.meta_.attributes
        }

    def clone(self):
        """Deepclone the command"""
        return copy.deepcopy(self)

    def _clone_with_values(self, **kwargs):
        """To be implemented in each command"""
        raise NotImplementedError



# Standard Library Imports
import re

from collections import defaultdict
from datetime import datetime

# Protean
from protean.core.aggregate import BaseAggregate
from protean.core.field.basic import DateTime, Integer, String
from protean.core.field.embedded import ValueObjectField
from protean.core.repository import BaseRepository
from protean.core.value_object import BaseValueObject


class Person(BaseAggregate):
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50, required=True)
    age = Integer(default=21)
    created_at = DateTime(default=datetime.now())


class PersonRepository(BaseRepository):
    pass


class Alien(BaseAggregate):
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50, required=True)
    age = Integer(default=21)


class User(BaseAggregate):
    email = String(max_length=255, required=True, unique=True)
    password = String(max_length=3026)


class Email(BaseValueObject):
    REGEXP = r"\"?([-a-zA-Z0-9.`?{}]+@\w+\.\w+)\"?"

    # This is the external facing data attribute
    address = String(max_length=254, required=True)

    def clean(self):
        """ Business rules of Email address """
        errors = defaultdict(list)

        if not bool(re.match(Email.REGEXP, self.address)):
            errors["address"].append("is invalid")

        return errors


class ComplexUser(BaseAggregate):
    email = ValueObjectField(Email, required=True)
    password = String(required=True, max_length=255)



"""Implementation of a dictionary based repository """

# Standard Library Imports
import copy
import json

from collections import defaultdict
from datetime import date, datetime
from itertools import count
from operator import itemgetter
from threading import Lock
from typing import Any
from uuid import UUID

# Protean
from protean.core.exceptions import ObjectNotFoundError
from protean.core.model import BaseModel
from protean.globals import current_uow
from protean.port.dao import BaseDAO, BaseLookup, ResultSet
from protean.port.provider import BaseProvider
from protean.utils import Database
from protean.utils.query import Q

# Global in-memory store of dict data. Keyed by name, to provide
# multiple named local memory caches.
_databases = defaultdict(dict)
_locks = defaultdict(Lock)
_counters = defaultdict(count)


def derive_schema_name(model_cls):
    if hasattr(model_cls.meta_, "schema_name"):
        return model_cls.meta_.schema_name
    else:
        return model_cls.meta_.entity_cls.meta_.schema_name


class MemoryModel(BaseModel):
    """A model for the dictionary repository"""

    @classmethod
    def from_entity(cls, entity) -> "MemoryModel":
        """Convert the entity to a dictionary record """
        dict_obj = {}
        for attribute_name in entity.meta_.attributes:
            dict_obj[attribute_name] = getattr(entity, attribute_name)
        return dict_obj

    @classmethod
    def to_entity(cls, item: "MemoryModel"):
        """Convert the dictionary record to an entity """
        return cls.meta_.entity_cls(item, raise_errors=False)


class MemorySession:
    def __init__(self, provider, new_connection=False):
        self._provider = provider
        self.is_active = True

        if (
            current_uow and self._provider.name in current_uow._sessions
        ) and not new_connection:
            self._db = current_uow._sessions[self._provider.name]._db
        else:
            self._db = {
                "data": copy.deepcopy(_databases),
                "lock": _locks.setdefault(self._provider.name, Lock()),
                "counters": _counters,
            }

    def add(self, element):
        if element.state_.is_persisted:
            dao = self._provider.get_dao(element.__class__)
            dao.update(element, element.to_dict())
        else:
            dao = self._provider.get_dao(element.__class__)
            dao.create(element.to_dict())

    def delete(self, element):
        dao = self._provider.get_dao(element.__class__)
        dao.delete(element)

    def commit(self):
        if current_uow and self._provider.name in current_uow._sessions:
            current_uow._sessions[self._provider.name]._db["data"] = self._db["data"]
        else:
            global _databases
            _databases = self._db["data"]

    def rollback(self):
        pass

    def close(self):
        pass


class MemoryProvider(BaseProvider):
    """Provider class for Dict Repositories"""

    def __init__(self, name, domain, conn_info: dict):
        """Initialize Provider with Connection/Adapter details"""

        # In case of `MemoryProvider`, the `DATABASE` value will always be `MEMORY`.
        conn_info["DATABASE"] = Database.MEMORY.value
        super().__init__(name, domain, conn_info)

        # A temporary cache of already constructed model classes
        self._model_classes = {}

    def get_session(self):
        """Return a session object

        For Dictionary Repo, a session translates to a copy of the
        `database`. All transactions on the Provider's repositories
        are committed on this copy of the database.
        """
        return MemorySession(self)

    def get_connection(self, session_cls=None):
        """Return the dictionary database object """
        return MemorySession(self, new_connection=True)

    def _data_reset(self):
        """Reset data"""
        global _databases, _locks, _counters
        _databases = defaultdict(dict)
        _locks = defaultdict(Lock)
        _counters = defaultdict(count)

        # Discard any active Unit of Work
        if current_uow and current_uow.in_progress:
            current_uow.rollback()

    def decorate_model_class(self, entity_cls, model_cls):
        schema_name = derive_schema_name(model_cls)

        # Return the model class if it was already seen/decorated
        if schema_name in self._model_classes:
            return self._model_classes[schema_name]

        # If `model_cls` is already subclassed from MemoryModel,
        #   this method call is a no-op
        if issubclass(model_cls, MemoryModel):
            return model_cls
        else:
            custom_attrs = {
                key: value
                for (key, value) in vars(model_cls).items()
                if key not in ["Meta", "__module__", "__doc__", "__weakref__"]
            }

            # Protean
            from protean.core.model import ModelMeta

            meta_ = ModelMeta()
            meta_.entity_cls = entity_cls

            custom_attrs.update({"meta_": meta_})
            # FIXME Ensure the custom model attributes are constructed properly
            decorated_model_cls = type(
                model_cls.__name__, (MemoryModel, model_cls), custom_attrs
            )

            # Memoize the constructed model class
            self._model_classes[schema_name] = decorated_model_cls

            return decorated_model_cls

    def construct_model_class(self, entity_cls):
        """Return associated, fully-baked Model class"""
        model_cls = None

        # Return the model class if it was already seen/decorated
        if entity_cls.meta_.schema_name in self._model_classes:
            model_cls = self._model_classes[entity_cls.meta_.schema_name]
        else:
            # Protean
            from protean.core.model import ModelMeta

            meta_ = ModelMeta()
            meta_.entity_cls = entity_cls

            attrs = {
                "meta_": meta_,
            }
            # FIXME Ensure the custom model attributes are constructed properly
            model_cls = type(entity_cls.__name__ + "Model", (MemoryModel,), attrs)

            # Memoize the constructed model class
            self._model_classes[entity_cls.meta_.schema_name] = model_cls

        # Set Entity Class as a class level attribute for the Model, to be able to reference later.
        return model_cls

    def get_dao(self, entity_cls, model_cls):
        """Return a DAO object configured with a live connection"""
        return DictDAO(self.domain, self, entity_cls, model_cls)

    def _evaluate_lookup(self, key, value, negated, db):
        """Extract values from DB that match the given criteria"""
        results = {}
        for record_key, record_value in db.items():
            match = True

            stripped_key, lookup_class = self._extract_lookup(key)
            lookup = lookup_class(record_value[stripped_key], value)

            if record_value[stripped_key] is not None:
                if negated:
                    match &= not eval(lookup.as_expression())
                else:
                    match &= eval(lookup.as_expression())
            else:
                match = False

            if match:
                results[record_key] = record_value

        return results

    def raw(self, query: Any, data: Any = None):
        """Run raw queries on the database

        As an example of running ``raw`` queries on a Dict repository, we will run the query
        on all possible schemas, and return all results.
        """
        assert isinstance(query, str)

        conn = self.get_connection()
        items = []

        for schema_name in conn._db["data"]:
            input_db = conn._db["data"][schema_name]
            try:
                # Ensures that the string contains double quotes around keys and values
                query = query.replace("'", '"')
                criteria = json.loads(query)

                for key, value in criteria.items():
                    input_db = self._evaluate_lookup(key, value, False, input_db)

                items.extend(list(input_db.values()))

            except json.JSONDecodeError:
                # FIXME Log Exception
                raise Exception("Query Malformed")
            except KeyError:
                # We encountered a repository where the key was not found
                pass

        return items


class DictDAO(BaseDAO):
    """A repository for storing data in a dictionary """

    def __repr__(self) -> str:
        return f"DictDAO <{self.entity_cls.__name__}>"

    def _set_auto_fields(self, model_obj):
        """Set the values of the auto field using counter"""
        conn = self._get_session()

        for field_name, field_obj in self.entity_cls.meta_.auto_fields:
            counter_key = f"{self.schema_name}_{field_name}"
            if not (field_name in model_obj and model_obj[field_name] is not None):
                # Increment the counter and it should start from 1
                counter = next(conn._db["counters"][counter_key])
                if not counter:
                    counter = next(conn._db["counters"][counter_key])

                model_obj[field_name] = counter

        return model_obj

    def _create(self, model_obj):
        """Write a record to the dict repository"""
        conn = self._get_session()

        # Update the value of the counters
        model_obj = self._set_auto_fields(model_obj)

        # Add the entity to the repository
        identifier = model_obj[self.entity_cls.meta_.id_field.field_name]
        with conn._db["lock"]:
            # Check if object is present
            if identifier in conn._db["data"][self.schema_name]:
                raise ObjectNotFoundError(
                    f"`{self.__class__.__name__}` object with identifier {identifier} "
                    f"already exists."
                )

            conn._db["data"][self.schema_name][identifier] = model_obj

        if not current_uow:
            conn.commit()
            conn.close()

        return model_obj

    def _filter_items(self, criteria: Q, db):
        """Recursive function to filter items from dictionary"""
        # Filter the dictionary objects based on the filters
        negated = criteria.negated
        input_db = None

        if criteria.connector == criteria.AND:
            # Trim database records over successive iterations
            #   Whatever is left at the end satisfy all criteria (AND)
            input_db = db
            for child in criteria.children:
                if isinstance(child, Q):
                    input_db = self._filter_items(child, input_db)
                else:
                    input_db = self.provider._evaluate_lookup(
                        child[0], child[1], negated, input_db
                    )
        else:
            # Grow database records over successive iterations
            #   Whatever is left at the end satisfy any criteria (OR)
            input_db = {}
            for child in criteria.children:
                if isinstance(child, Q):
                    results = self._filter_items(child, db)
                else:
                    results = self.provider._evaluate_lookup(
                        child[0], child[1], negated, db
                    )

                input_db = {**input_db, **results}

        return input_db

    def _filter(
        self, criteria: Q, offset: int = 0, limit: int = 10, order_by: list = ()
    ):
        """Read the repository and return results as per the filer"""
        conn = self._get_session()

        if criteria.children:
            items = list(
                self._filter_items(
                    criteria, conn._db["data"][self.schema_name]
                ).values()
            )
        else:
            items = list(conn._db["data"][self.schema_name].values())

        # Sort the filtered results based on the order_by clause
        for o_key in order_by:
            reverse = False
            if o_key.startswith("-"):
                reverse = True
                o_key = o_key[1:]
            items = sorted(items, key=itemgetter(o_key), reverse=reverse)

        result = ResultSet(
            offset=offset,
            limit=limit,
            total=len(items),
            items=items[offset : offset + limit],
        )

        return result

    def _update(self, model_obj):
        """Update the entity record in the dictionary """
        conn = self._get_session()

        identifier = model_obj[self.entity_cls.meta_.id_field.field_name]
        with conn._db["lock"]:
            # Check if object is present
            if identifier not in conn._db["data"][self.schema_name]:
                raise ObjectNotFoundError(
                    f"`{self.__class__.__name__}` object with identifier {identifier} "
                    f"does not exist."
                )

            conn._db["data"][self.schema_name][identifier] = model_obj

        if not current_uow:
            conn.commit()
            conn.close()

        return model_obj

    def _update_all(self, criteria: Q, *args, **kwargs):
        """Update all objects satisfying the criteria """
        conn = self._get_session()

        items = self._filter_items(criteria, conn._db["data"][self.schema_name])

        update_count = 0
        for key in items:
            item = items[key]
            item.update(*args)
            item.update(kwargs)
            conn._db["data"][self.schema_name][key] = item

            update_count += 1

        if not current_uow:
            conn.commit()
            conn.close()

        return update_count

    def _delete(self, model_obj):
        """Delete the entity record in the dictionary """
        conn = self._get_session()

        identifier = model_obj[self.entity_cls.meta_.id_field.field_name]
        with conn._db["lock"]:
            # Check if object is present
            if identifier not in conn._db["data"][self.schema_name]:
                raise ObjectNotFoundError(
                    f"`{self.entity_cls.__name__}` object with identifier {identifier} "
                    f"does not exist."
                )

            del conn._db["data"][self.schema_name][identifier]

        if not current_uow:
            conn.commit()
            conn.close()

        return model_obj

    def _delete_all(self, criteria: Q = None):
        """Delete the dictionary object by its criteria"""
        conn = self._get_session()
        items = []

        if criteria:
            # Delete the object from the dictionary and return the deletion count
            items = self._filter_items(criteria, conn._db["data"][self.schema_name])

            # Delete all the matching identifiers
            with conn._db["lock"]:
                for identifier in items:
                    conn._db["data"][self.schema_name].pop(identifier, None)
        else:
            with conn._db["lock"]:
                if self.schema_name in conn._db["data"]:
                    del conn._db["data"][self.schema_name]

        if not current_uow:
            conn.commit()
            conn.close()

        return len(items)

    def _raw(self, query: Any, data: Any = None):
        """Run raw query on Repository.

        For this stand-in repository, the query string is a json string that contains kwargs
        criteria with straight-forward equality checks. Individual criteria are always AND-ed
        and the result is always a subset of the full repository.

        We will ignore the `data` parameter for this kind of repository.
        """
        # Ensure that we are dealing with a string, for this repository
        assert isinstance(query, str)

        conn = self._get_session()
        input_db = conn._db["data"][self.schema_name]
        result = None

        try:
            # Ensures that the string contains double quotes around keys and values
            query = query.replace("'", '"')
            criteria = json.loads(query)

            for key, value in criteria.items():
                input_db = self.provider._evaluate_lookup(key, value, False, input_db)

            items = list(input_db.values())
            result = ResultSet(
                offset=1, limit=len(items), total=len(items), items=items
            )

        except json.JSONDecodeError:
            # FIXME Log Exception
            raise Exception("Query Malformed")

        if not current_uow:
            conn.commit()
            conn.close()

        return result


operators = {
    "exact": "==",
    "iexact": "==",
    "contains": "in",
    "icontains": "in",
    "gt": ">",
    "gte": ">= ",
    "lt": "<",
    "lte": "<=",
    "in": "in",
}


class MemoryLookup(BaseLookup):
    """Base class with default implementation of expression construction"""

    def process_source(self):
        """Return source with transformations, if any"""
        if isinstance(self.source, (UUID, datetime, date)):
            self.source = str(self.source)

        if isinstance(self.source, str):
            # Replace single and double quotes with escaped single-quote
            self.source = self.source.replace("'", "'").replace('"', "'")
            return '"{source}"'.format(source=self.source)
        return self.source

    def process_target(self):
        """Return target with transformations, if any"""
        if isinstance(self.target, (UUID, datetime, date)):
            self.target = str(self.target)

        if isinstance(self.target, str):
            # Replace single and double quotes with escaped single-quote
            self.target = self.target.replace("'", "'").replace('"', "'")
            return '"{target}"'.format(target=self.target)
        return self.target

    def as_expression(self):
        return "{source} {op} {target}".format(
            source=self.process_source(),
            op=operators[self.lookup_name],
            target=self.process_target(),
        )


@MemoryProvider.register_lookup
class Exact(MemoryLookup):
    """Exact Match Query"""

    lookup_name = "exact"


@MemoryProvider.register_lookup
class IExact(MemoryLookup):
    """Exact Case-Insensitive Match Query"""

    lookup_name = "iexact"

    def process_source(self):
        """Return source in lowercase"""
        assert isinstance(self.source, str)
        return "%s.lower()" % super().process_source()

    def process_target(self):
        """Return target in lowercase"""
        assert isinstance(self.target, str)
        return "%s.lower()" % super().process_target()


@MemoryProvider.register_lookup
class Contains(MemoryLookup):
    """Exact Contains Query"""

    lookup_name = "contains"

    def as_expression(self):
        """Check for Target string to be in Source string"""
        return "%s %s %s" % (
            self.process_target(),
            operators[self.lookup_name],
            self.process_source(),
        )


@MemoryProvider.register_lookup
class IContains(MemoryLookup):
    """Exact Case-Insensitive Contains Query"""

    lookup_name = "icontains"

    def process_source(self):
        """Return source in lowercase"""
        assert isinstance(self.source, str)
        return "%s.lower()" % super().process_source()

    def process_target(self):
        """Return target in lowercase"""
        assert isinstance(self.target, str)
        return "%s.lower()" % super().process_target()

    def as_expression(self):
        """Check for Target string to be in Source string"""
        return "%s %s %s" % (
            self.process_target(),
            operators[self.lookup_name],
            self.process_source(),
        )


@MemoryProvider.register_lookup
class GreaterThan(MemoryLookup):
    """Greater than Query"""

    lookup_name = "gt"


@MemoryProvider.register_lookup
class GreaterThanOrEqual(MemoryLookup):
    """Greater than or Equal Query"""

    lookup_name = "gte"


@MemoryProvider.register_lookup
class LessThan(MemoryLookup):
    """Less than Query"""

    lookup_name = "lt"


@MemoryProvider.register_lookup
class LessThanOrEqual(MemoryLookup):
    """Less than or Equal Query"""

    lookup_name = "lte"


@MemoryProvider.register_lookup
class In(MemoryLookup):
    """In Query"""

    lookup_name = "in"

    def process_target(self):
        """Ensure target is a list or tuple"""
        assert type(self.target) in (list, tuple)
        return super().process_target()



""" Module for defining basic Field types of Entity """

# Standard Library Imports
import datetime

# Protean
from dateutil.parser import parse as date_parser
from protean.core.exceptions import InvalidOperationError, ValidationError
from protean.core.field import validators
from protean.core.field.base import Field


class String(Field):
    """Concrete field implementation for the string type.

    :param max_length: The maximum allowed length for the field.
    :param min_length: The minimum allowed length for the field.

    FIXME Should max_length be optional?
    """

    default_error_messages = {
        "invalid": '{value}" value must be a string.',
    }

    def __init__(self, max_length=255, min_length=None, **kwargs):
        self.min_length = min_length
        self.max_length = max_length
        self.default_validators = [
            validators.MinLengthValidator(self.min_length),
            validators.MaxLengthValidator(self.max_length),
        ]
        super().__init__(**kwargs)

    def _cast_to_type(self, value):
        """ Convert the value to its string representation"""
        if isinstance(value, str) or value is None:
            return value
        return str(value)


class Text(Field):
    """Concrete field implementation for the text type.
    """

    default_error_messages = {
        "invalid": '{value}" value must be a string.',
    }

    def _cast_to_type(self, value):
        """ Convert the value to its string representation"""
        if isinstance(value, str) or value is None:
            return value
        return str(value)


class Integer(Field):
    """Concrete field implementation for the Integer type.

    :param min_value: The minimum allowed value for the field.
    :param max_value: The maximum allowed value for the field.

    """

    default_error_messages = {
        "invalid": '"{value}" value must be an integer.',
    }

    def __init__(self, min_value=None, max_value=None, **kwargs):
        self.min_value = min_value
        self.max_value = max_value
        self.default_validators = [
            validators.MinValueValidator(self.min_value),
            validators.MaxValueValidator(self.max_value),
        ]
        super().__init__(**kwargs)

    def _cast_to_type(self, value):
        """ Convert the value to an int and raise error on failures"""
        try:
            if isinstance(value, str):
                if value.strip() == "":
                    return None
            return int(value)
        except (ValueError, TypeError):
            self.fail("invalid", value=value)


class Float(Field):
    """Concrete field implementation for the Floating type.

    :param min_value: The minimum allowed value for the field.
    :param max_value: The maximum allowed value for the field.

    """

    default_error_messages = {
        "invalid": '"{value}" value must be floating point number.',
    }

    def __init__(self, min_value=None, max_value=None, **kwargs):
        self.min_value = min_value
        self.max_value = max_value
        self.default_validators = [
            validators.MinValueValidator(self.min_value),
            validators.MaxValueValidator(self.max_value),
        ]
        super().__init__(**kwargs)

    def _cast_to_type(self, value):
        """ Convert the value to a float and raise error on failures"""
        try:
            return float(value)
        except (ValueError, TypeError):
            self.fail("invalid", value=value)


class Boolean(Field):
    """Concrete field implementation for the Boolean type.
    """

    default_error_messages = {
        "invalid": '"{value}" value must be either True or False.',
    }

    def _cast_to_type(self, value):
        """ Convert the value to a boolean and raise error on failures"""
        if value in (True, False):
            return bool(value)
        if value in ("t", "True", "1"):
            return True
        if value in ("f", "False", "0"):
            return False
        self.fail("invalid", value=value)


class List(Field):
    """Concrete field implementation for the List type.
    """

    default_error_messages = {
        "invalid": '"{value}" value must be of list type.',
        "invalid_content": "Invalid value",
    }

    def __init__(self, content_type=Text, pickled=False, **kwargs):
        if content_type not in [
            Boolean,
            Date,
            DateTime,
            Float,
            Identifier,
            Integer,
            String,
            Text,
        ]:
            raise ValidationError({"content_type": ["Content type not supported"]})
        self.content_type = content_type
        self.pickled = pickled

        super().__init__(**kwargs)

    def _cast_to_type(self, value):
        """ Raise errors if the value is not a list, or
        the items in the list are not of the right data type.
        """
        if not isinstance(value, list):
            self.fail("invalid", value=value)

        # Try to cast value into the destination type.
        #   Throw error if the underlying type does not support value.
        new_value = []
        try:
            for item in value:
                new_value.append(self.content_type()._load(item))
        except ValidationError:
            self.fail("invalid_content", value=value)

        if new_value != value:
            self.fail("invalid_content", value=value)

        return value


class Set(Field):
    """Concrete field implementation for the Set type.
    """

    default_error_messages = {
        "invalid": '"{value}" value must be of set type.',
    }

    def _cast_to_type(self, value):
        """ Raise error if the value is not a set """
        if not isinstance(value, set):
            self.fail("invalid", value=value)
        return value


class Dict(Field):
    """Concrete field implementation for the Dict type.
    """

    default_error_messages = {
        "invalid": '"{value}" value must be of dict type.',
    }

    def __init__(self, pickled=False, **kwargs):
        self.pickled = pickled

        super().__init__(**kwargs)

    def _cast_to_type(self, value):
        """ Raise error if the value is not a dict """

        # Lists are allowed in JSON because Postgres supports them.
        #   All other databases treat these columns as BLOB/Pickled Type.
        if not isinstance(value, (dict, list)):
            self.fail("invalid", value=value)
        return value


class Auto(Field):
    """ Concrete field implementation for the Database Autogenerated types.
    """

    def __init__(self, *args, **kwargs):
        """Initialize an Auto Field"""
        # Force set required to false
        super().__init__(*args, **kwargs)
        self.required = False

    def __set__(self, instance, value):
        """An Identifier Field once set cannot be reset or changed.
        We override the ``__set__`` method and prevent setting of new value if one was
        already set and is different from the new value
        """
        existing_value = getattr(instance, self.field_name)
        if existing_value is not None and value != existing_value:
            raise InvalidOperationError("Identifiers cannot be changed once set")

        value = self._load(value)
        instance.__dict__[self.field_name] = value

        if hasattr(instance, "state_"):
            # Mark Entity as Dirty
            instance.state_.mark_changed()

    def _cast_to_type(self, value):
        """ Perform no validation for auto fields. Return the value as is"""
        return value


class Identifier(Field):
    """ Concrete field implementation for Identifiers.

    Values can be Integers or Strings.
    """

    def _cast_to_type(self, value):
        """ Perform no validation for identifier fields. Return the value as is"""
        return value


class CustomObject(Field):
    """Concrete field implementation for objects. Values are instances of custom classes"""

    def _cast_to_type(self, value):
        """ Perform no validation for identifier fields. Return the value as is"""
        return value


class Method(Field):
    """Helper field for custom methods associated with serializer fields"""

    def __init__(self, method_name, **kwargs):
        self.method_name = method_name
        super().__init__(**kwargs)

    def _cast_to_type(self, value):
        """ Perform no validation for identifier fields. Return the value as is"""
        return value


class Nested(Field):
    """Helper field for nested objects associated with serializer fields"""

    def __init__(self, schema_name, many=False, **kwargs):
        self.schema_name = schema_name
        self.many = many
        super().__init__(**kwargs)

    def _cast_to_type(self, value):
        """ Perform no validation for identifier fields. Return the value as is"""
        return value


class Date(Field):
    """ Concrete field implementation for the Date type.
    """

    default_error_messages = {
        "invalid": '"{value}" has an invalid date format.',
    }

    def _cast_to_type(self, value):
        """ Convert the value to a date and raise error on failures"""
        if isinstance(value, datetime.datetime):
            return value.date()
        if isinstance(value, datetime.date):
            return value
        try:
            value = date_parser(value)
            return value.date()
        except ValueError:
            self.fail("invalid", value=value)


class DateTime(Field):
    """ Concrete field implementation for the Datetime/Timestamp type.
    """

    def _cast_to_type(self, value):
        """ Convert the value to a datetime and raise error on failures"""
        if isinstance(value, datetime.datetime):
            return value
        if isinstance(value, datetime.date):
            value = datetime.datetime(value.year, value.month, value.day)
            return value
        try:
            value = date_parser(value)
            return value
        except ValueError:
            self.fail("invalid", value=value)



"""Module for defining embedded fields"""

# Protean
from protean.core.field.base import Field

# Local/Relative Imports
from .utils import (
    fetch_aggregate_cls_from_domain,
    fetch_entity_cls_from_domain,
    fetch_value_object_cls_from_domain,
)


class _ShadowField(Field):
    """Shadow Attribute Field to back Value Object Fields"""

    def __init__(self, owner, field_name, field_type, **kwargs):
        """Preserve link to owner, and original field type for later reference"""
        super().__init__(**kwargs)

        self.owner = owner
        self.field_name = field_name
        self.field_type = field_type

    def __set__(self, instance, value):
        """Override `__set__` to update owner field and silently fail to update values.
        When the value object's value is set, the embedded fields will be automatically filled.
        """
        pass

    def __delete__(self, instance):
        """Nullify values and linkages"""
        self._reset_values(instance)

    def _cast_to_type(self, value):
        """Verify type of value assigned to the shadow field"""
        # FIXME Verify that the value being assigned is compatible with the remote field
        return value

    def _reset_values(self, instance):
        """Reset all associated values and clean up dictionary items"""
        instance.__dict__.pop(self.field_name, None)


class ValueObjectField(Field):
    """Field implementation for Value Objects"""

    def __init__(self, value_object_cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._value_object_cls = value_object_cls

        self.embedded_fields = {}
        for (
            field_name,
            field_obj,
        ) in self._value_object_cls.meta_.declared_fields.items():
            self.embedded_fields[field_name] = _ShadowField(
                self,
                field_name,
                field_obj.__class__,
                # FIXME Pass all other kwargs here
                #   Because we want the shadow field to mimic the behavior of the actual field
                #   Which means that ShadowField somehow has to become an Integer, Float, String, etc.
                referenced_as=field_obj.referenced_as,
            )

    @property
    def value_object_cls(self):
        """Property to retrieve value_object_cls as a Value Object when possible"""
        # Checks if ``value_object_cls`` is a string
        #   If it is, checks if the Value Object is imported and available
        #   If it is, register the class
        try:
            if isinstance(self._value_object_cls, str):
                self._value_object_cls = fetch_value_object_cls_from_domain(
                    self._value_object_cls
                )
        except AssertionError:
            # Preserve ``value_object_cls`` as a string and we will hook up the entity later
            pass

        return self._value_object_cls

    def __set_name__(self, entity_cls, name):
        super().__set_name__(entity_cls, name)
        # Refresh underlying embedded field names
        for embedded_field in self.embedded_fields.values():
            if embedded_field.referenced_as:
                embedded_field.attribute_name = embedded_field.referenced_as
            else:
                embedded_field.attribute_name = (
                    self.field_name + "_" + embedded_field.field_name
                )

    def get_shadow_fields(self):
        """Return shadow field
        Primarily used during Entity initialization to register shadow field"""
        shadow_fields = []
        for field in self.embedded_fields.values():
            shadow_fields.append((field.attribute_name, field))
        return shadow_fields

    def _cast_to_type(self, value):
        if not isinstance(value, self._value_object_cls):
            self.fail("invalid", value=value)
        return value

    def __set__(self, instance, value):
        """Override `__set__` to coordinate between value object and its embedded fields"""
        if isinstance(self.value_object_cls, str):
            self.value_object_cls = fetch_value_object_cls_from_domain(
                self.value_object_cls
            )

            # Refresh attribute name, now that we know `value_object_cls` class
            self.attribute_name = self.get_attribute_name()

        value = self._load(value)

        if value:
            # Check if the reference object has been saved. Otherwise, throw ValueError
            self._set_own_value(instance, value)
            self._set_embedded_values(instance, value)
        else:
            self._reset_values(instance)

    def _set_own_value(self, instance, value):
        if value is None:
            instance.__dict__.pop(self.field_name, None)
        else:
            instance.__dict__[self.field_name] = value

    def _set_embedded_values(self, instance, value):
        if value is None:
            for field_name in self.embedded_fields:
                attribute_name = self.embedded_fields[field_name].attribute_name
                instance.__dict__.pop(attribute_name, None)
                self.embedded_fields[field_name].value = None
        else:
            for field_name in value.meta_.declared_fields:
                self.embedded_fields[field_name].value = getattr(value, field_name)
                attribute_name = self.embedded_fields[field_name].attribute_name
                instance.__dict__[attribute_name] = getattr(value, field_name)

    def __delete__(self, instance):
        self._reset_values(instance)

    def _reset_values(self, instance):
        """Reset all associated values and clean up dictionary items"""
        self._set_own_value(instance, None)
        self._set_embedded_values(instance, None)


class AggregateField(Field):
    """Field implementation for Aggregates"""

    def __init__(self, aggregate_cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._aggregate_cls = aggregate_cls

    @property
    def aggregate_cls(self):
        """Property to retrieve aggregate_cls as a Aggregate when possible"""
        # Checks if ``aggregate_cls`` is a string
        #   If it is, checks if the Aggregate is imported and available
        #   If it is, register the class
        try:
            if isinstance(self._aggregate_cls, str):
                self._aggregate_cls = fetch_aggregate_cls_from_domain(
                    self._aggregate_cls
                )
        except AssertionError:
            # Preserve ``aggregate_cls`` as a string and we will hook up the entity later
            pass

        return self._aggregate_cls

    def __get__(self, instance, owner):
        if hasattr(instance, "__dict__"):
            return instance.__dict__.get(self.field_name, self.value).to_dict()

    def __set_name__(self, element_cls, name):
        super().__set_name__(element_cls, name)

    def _cast_to_type(self, value):
        if not isinstance(value, self._aggregate_cls):
            self.fail("invalid", value=value)
        return value


class EntityField(Field):
    """Field implementation for Entities"""

    def __init__(self, entity_cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._entity_cls = entity_cls

    @property
    def entity_cls(self):
        """Property to retrieve entity_cls as a Entity when possible"""
        # Checks if ``entity_cls`` is a string
        #   If it is, checks if the Aggregate is imported and available
        #   If it is, register the class
        try:
            if isinstance(self._entity_cls, str):
                self._entity_cls = fetch_entity_cls_from_domain(self._entity_cls)
        except AssertionError:
            # Preserve ``entity_cls`` as a string and we will hook up the entity later
            pass

        return self._entity_cls

    def __set_name__(self, element_cls, name):
        super().__set_name__(element_cls, name)

    def _cast_to_type(self, value):
        if not isinstance(value, self._entity_cls):
            self.fail("invalid", value=value)
        return value



# Local/Relative Imports
from .aggregate_elements_with_value_objects import File, Resource


class TestAggregatesWithValueObjects:
    def test_that_value_object_can_be_embedded_within_aggregate(self):
        resource = Resource(
            title="Resource 1", associated_file=File(url="/server/1.pdf", type="PDF")
        )

        assert resource is not None

    def test_that_embedded_value_objects_are_unique(self):
        resource1 = Resource(
            title="Resource 1", associated_file=File(url="/server/1.pdf", type="PDF")
        )
        resource2 = Resource(
            title="Resource 2", associated_file=File(url="/server/2.pdf", type="PDF")
        )

        assert resource1.associated_file is not None
        assert resource2.associated_file is not None
        assert resource1.associated_file != resource2.associated_file
        assert resource1.associated_file.url == "/server/1.pdf"
        assert resource1.associated_file_url == "/server/1.pdf"
        assert resource2.associated_file.url == "/server/2.pdf"
        assert resource2.associated_file_url == "/server/2.pdf"

    def test_that_attributes_from_embedded_objects_are_attached(self):
        resource1 = Resource(
            title="Resource 1", associated_file=File(url="/server/1.pdf", type="PDF")
        )
        assert all(
            key in resource1.__dict__
            for key in [
                "title",
                "associated_file",
                "associated_file_url",
                "associated_file_type",
            ]
        )

        resource2 = Resource(title="Resource 2")
        assert all(
            key in resource2.__dict__
            for key in ["title", "associated_file_url", "associated_file_type"]
        )



# Standard Library Imports
from datetime import datetime

# Protean
from protean.adapters.repository.memory import Exact


class TestLookup:
    """This class holds tests for Lookup Class"""

    # Protean
    from protean.adapters.repository.memory import MemoryProvider
    from protean.port.dao import BaseLookup

    @MemoryProvider.register_lookup
    class SampleLookup(BaseLookup):
        """A simple implementation of lookup class"""

        lookup_name = "sample"

        def as_expression(self):
            return "%s %s %s" % (self.process_source(), "<<<>>>", self.process_target())

    def test_initialization_of_a_lookup_object(self):
        lookup = self.SampleLookup("src", "trg")
        assert lookup.as_expression() == "src <<<>>> trg"

    def test_registration_of_a_lookup_to_an_adapter(self):
        # Protean
        from protean.adapters.repository.memory import MemoryProvider

        assert MemoryProvider.get_lookups().get("sample") == self.SampleLookup

    def test_expression_constructed_for_datetime_fields(self):
        now = datetime.today()
        lookup = Exact("datetimefield", now)
        assert lookup is not None
        assert lookup.as_expression() == f'"datetimefield" == "{str(now)}"'

    def test_expression_constructed_for_date_fields(self):
        today = datetime.today().date()
        lookup = Exact("datefield", today)
        assert lookup is not None
        assert lookup.as_expression() == f'"datefield" == "{str(today)}"'
